#pragma once
#include "tendes_plane.h"
#include "PaintingStyleField.h"
#include <string>


class EdgeFieldManipulator {
private:
	double weight_interpolate;
	double current_rotation_radian;
	double set_rotation_degree;
	int init_edgefld_option;

	PaintingStyleField *psf;
	Polyhedron *poly;


	int determine_within_box(icVector2 center, icVector2 min_dir, icVector2 max_dir, double min_mag, double max_mag, icVector2 pos);
	void interpolateField_noAnalyze(int time, int region_id = -1, int set_zero = 0);


public:
	EdgeFieldManipulator();

	void setMesh(Polyhedron &poly);
	void releaseMesh();
	bool isMeshSet() { return (poly != NULL) ? true : false; }

	void setStyleField(PaintingStyleField &psf);
	void releaseStyleField();
	bool isStyleFieldSet() { return (psf != NULL) ? true : false; }

	void setWeight(double weight);
	double getWeight();

	void setRotationAngle(double rot_degree);
	double getRotationAngle();

	void smoothField(int num_iterations, double strong_vertex_ratio);
	void normalizeField();

	int addSink(double x, double y, int time = 0, int region_id = -1);
	int addSource(double x, double y, int time = 0, int region_id = -1);
	int addSaddle(double x, double y, int time = 0, int region_id = -1);
	int addCCWCenter(double x, double y, int time = 0, int region_id = -1);
	int addCWCenter(double x, double y, int time = 0, int region_id = -1);
	int addDipole1(double x, double y, int time = 0, int region_id = -1);
	int addDipole2(double x, double y, int time = 0, int region_id = -1);
	int addMonkeySaddle(double x, double y, int time = 0, int region_id = -1);
	void editSE(double x, double y, int se_id, int edit_mode, int time);

	int addRegularElement(double x, double y, int time = 0, int region_id = -1);
	void editRE(double x, double y, int re_id, int edit_mode, int time);

	void selectElement(double x, double y, int &se_id, int &re_id, int time);
	void selectAnalyticalElement(double x, double y, int &s_id);
	void determineSEEditMode(double x, double y, int &se_id, int &edit_mode);
	void determineREEditMode(double x, double y, int &re_id, int &edit_mode);

	bool removeElement(double x, double y, int time = 0);

	bool cancelSingularityPair(int s_id1, int s_id2);
	bool moveSingularity(int s_id, int t_id);

	void rotateField();
	void reflectField();

	void undo();
	void redo();

	void setDesignElements(SingularElementList *selist, RegularElementList *relist, int time);
	void getDesignElements(SingularElementList *&selist, RegularElementList *&relist);

	void smoothRegionInField();
	void interpolateField(int time, int region_id = -1);
	void resetField(int time);
	void zeroField(int time, int region_id = -1);

	void clearSelectedRegion();
	void updateSelectedRegion(int selected_t_id);

	void setInitEdgeFieldOption(int mode);							// 0: edge field from image, 1: zero edge field
	int setInitEdgeFieldOption() { return init_edgefld_option; }

	bool replaceField(string &filename);
	bool exportField(string &filename);
};

