#pragma once
#include "StdAfx.h"
#include "EdgeFieldSampler.h"

#define USE_OMP 1

#ifndef _OPENMP
#define USE_OMP 0
#endif

#if USE_OMP == 1
#include <omp.h>
#endif


inline icVector2 EdgeFieldSampler::sample_pixel_by_xy_vector(Polyhedron &poly, double x, double y, icVector2 q[3], Triangle *temp_t, double t_a, int *w, int *h) {
	icVector2 result;
	double a[3];

	*w = (int)(x / num_sample);
	*h = (int)(y / num_sample);
	icVector2 new_q(x, y);
	a[0]=poly.calc_triangle_area(new_q, q[1], q[2])/t_a;
	a[1]=poly.calc_triangle_area(new_q, q[2], q[0])/t_a;
	a[2]=1.0-a[0]-a[1]; //calc_triangle_area(new_q, q[0], q[1])/t_a;
	result=a[0]*temp_t->verts[0]->vec_fld 
				+a[1]*temp_t->verts[1]->vec_fld 
				+a[2]*temp_t->verts[2]->vec_fld;
	return result;
}


inline icMatrix2x2 EdgeFieldSampler::sample_pixel_by_xy_tensor(Polyhedron &poly, double x, double y, icVector2 q[3], Triangle *temp_t, double t_a, int *w, int *h) {
	icMatrix2x2 result;
	double a[3];

	*w = (int)(x / num_sample);
	*h = (int)(y / num_sample);
	icVector2 new_q(x, y);
	a[0]=poly.calc_triangle_area(new_q, q[1], q[2])/t_a;
	a[1]=poly.calc_triangle_area(new_q, q[2], q[0])/t_a;
	a[2]=1.0-a[0]-a[1]; //calc_triangle_area(new_q, q[0], q[1])/t_a;
	result=a[0]*temp_t->verts[0]->ten_fld 
				+a[1]*temp_t->verts[1]->ten_fld 
				+a[2]*temp_t->verts[2]->ten_fld;
	return result;
}


void EdgeFieldSampler::sampleToMesh(Polyhedron &poly, FieldSampleImage &fsi) {
	if(poly.vec_ten_mode == 0) {
		sampleVecFieldToMesh(poly, fsi);
	}
	else if(poly.vec_ten_mode == 1) {
		sampleTenFieldToMesh(poly, fsi);
	}
}


void EdgeFieldSampler::sampleToFSI(Polyhedron &poly, FieldSampleImage &fsi) {
	if(poly.vec_ten_mode == 0) {
		sampleVecFieldToFSI(poly, fsi);
	}
	else if(poly.vec_ten_mode == 1) {
		sampleTenFieldToFSI(poly, fsi);
	}
}


void EdgeFieldSampler::sampleVecFieldToMesh(Polyhedron &poly, FieldSampleImage &fsi) {
	int width, height, num_verts;
	int wx, wy, wx_l, wy_u, total_sample;
	int w_1, h_1;
	Vertex *v;
	icVector2 tmp;

	num_verts = poly.nverts;
	width = fsi.w;
	height = fsi.h;
	w_1 = width - 1;
	h_1 = height - 1;

	//w_1++;h_1++;

	for(int i = 0; i < num_verts; i++) {
		v = poly.vlist[i];

		wx = (int)((v->x + 0.5) * w_1);
		wy = (int)((0.5 - v->y) * h_1);

		wx_l = wx + 1;
		wy_u = wy + 1;
		total_sample = 0;
		tmp.set(0, 0);

		if(wx >= 0 && wx < width) {
			if(wy >= 0 && wy < height) {
				tmp += fsi.fld[wy * width + wx].vec_fld;
				total_sample++;
			}

			if(wy_u >= 0 && wy_u < height) {
				tmp += fsi.fld[wy_u * width + wx].vec_fld;
				total_sample++;
			}
		}

		if(wx_l >= 0 && wx_l < width) {
			if(wy >= 0 && wy < height) {
				tmp += fsi.fld[wy * width + wx_l].vec_fld;
				total_sample++;
			}

			if(wy_u >= 0 && wy_u < height) {
				tmp += fsi.fld[wy_u * width + wx_l].vec_fld;
				total_sample++;
			}
		} 


		if(total_sample > 0) {
			tmp.entry[0] /= total_sample;
			tmp.entry[1] /= total_sample;
		}

		v->vec_fld = tmp;
	}
}


void EdgeFieldSampler::sampleTenFieldToMesh(Polyhedron &poly, FieldSampleImage &fsi) {
	int width, height, num_verts;
	int wx, wy, wx_l, wy_u, total_sample;
	int w_1, h_1;
	double angle, this_len;
	Vertex *v;
	icMatrix2x2 tmp;

	num_verts = poly.nverts;
	width = fsi.w;
	height = fsi.h;
	w_1 = width - 1;
	h_1 = height - 1;

	//w_1++;h_1++;

	for(int i = 0; i < num_verts; i++) {
		v = poly.vlist[i];

		tmp.set(0.0);
		wx = (int)((v->x + 0.5) * w_1);
		wy = (int)((0.5 - v->y) * h_1);

		wx_l = wx + 1;
		wy_u = wy + 1;
		total_sample = 0;

		if(wx >= 0 && wx < width) {
			if(wy >= 0 && wy < height) {
				tmp += fsi.fld[wy * width + wx].ten_fld;
				total_sample++;
			}

			if(wy_u >= 0 && wy_u < height) {
				tmp += fsi.fld[wy_u * width + wx].ten_fld;
				total_sample++;
			}
		}

		if(wx_l >= 0 && wx_l < width) {
			if(wy >= 0 && wy < height) {
				tmp += fsi.fld[wy * width + wx_l].ten_fld;
				total_sample++;
			}

			if(wy_u >= 0 && wy_u < height) {
				tmp += fsi.fld[wy_u * width + wx_l].ten_fld;
				total_sample++;
			}
		}


		if(total_sample > 0) {
			tmp.entry[0][0] /= total_sample;
			tmp.entry[0][1] /= total_sample;
			tmp.entry[1][0] /= total_sample;
			tmp.entry[1][1] /= total_sample;
			this_len = length(icVector2(tmp.entry[0][0], tmp.entry[0][1]));
			angle = atan2(tmp.entry[0][1], tmp.entry[0][0]) * 0.5;
			v->vec_fld.set(this_len * cos(angle), this_len * sin(angle));
		}
		else {
			v->vec_fld.set(0, 0);
		}

		v->ten_fld = tmp;
	}
}


void EdgeFieldSampler::sampleVecFieldToFSI(Polyhedron &poly, FieldSampleImage &fsi) {
	int i;
	int index;
	double l_x, r_x, d_l, d_r;
	double lowest, highest, x, y, t_a;
	int n_horizontal=0;
	int low, middle, up;
	icVector2 q[3];
	int w, h;
	int sw, sh;
	int middle_on_right;
	icVector2 new_vec;
	unsigned char *handled;
	int num_pixels;


	num_pixels = fsi.w * fsi.h;

	sw = fsi.w * num_sample;
	sh = fsi.h * num_sample;

	handled = (unsigned char *)malloc(sizeof(unsigned char) * num_pixels);
	memset(handled, 0, sizeof(unsigned char) * num_pixels);

#if USE_OMP == 1
	#pragma omp parallel private(i, index, l_x, r_x, d_l, d_r, lowest, highest, x, y, t_a, low, middle, up, q, w, h, middle_on_right, new_vec)
	{
#endif


#if USE_OMP == 1
	#pragma omp for
#endif
	for(i = 0; i < num_pixels; i++) {
		fsi.fld[i].vec_fld.set(0.0);
	}


#if USE_OMP == 1
	#pragma omp for
#endif
	for(i = 0; i < poly.ntris; i++) {
		Triangle *temp_t=poly.tlist[i];
		Vertex *v[3];

		for(int j = 0; j < 3; j++) v[j] = temp_t->verts[j];

		if(fabs(v[0]->x) > 0.5 && fabs(v[0]->y) > 0.5 && fabs(v[1]->x) > 0.5 && fabs(v[1]->y) > 0.5 && fabs(v[2]->x) > 0.5 && fabs(v[2]->y) > 0.5)
			continue;

		for(int j = 0; j < 3; j++) q[j].set((v[j]->x + 0.5) * sw, (0.5 - v[j]->y) * sh);

		t_a = poly.calc_triangle_area(q[0], q[1], q[2]);

		low =0;
		up = 0;
		lowest=q[0].entry[1];
		highest=q[0].entry[1];
		if (lowest>q[1].entry[1]) {
			lowest = q[1].entry[1];
			low = 1;
		}
		if (lowest>q[2].entry[1]) {
			lowest = q[2].entry[1];
			low = 2;
		}
		if (highest<q[1].entry[1]) {
			highest = q[1].entry[1];
			up = 1;
		}
		if (highest<q[2].entry[1]) {
			highest = q[2].entry[1];
			up = 2;
		}
		if (low !=0 && up != 0)
			middle =0;
		if (low !=1 && up != 1)
			middle =1;
		if (low !=2 && up != 2)
			middle =2;
	
		if (q[low].entry[1] == q[middle].entry[1]) { // bottom horizontal
			if (q[low].entry[0] > q[middle].entry[0]) {
				int temp = low;
				low = middle;
				middle = temp;
			}
			d_l = (q[up].entry[0]-q[low].entry[0])/(q[up].entry[1]-q[low].entry[1]);
			d_r = (q[up].entry[0]-q[middle].entry[0])/(q[up].entry[1]-q[middle].entry[1]);
			
			double y_min = q[low].entry[1];
			if (y_min == ceil(y_min)) {  // special case, y_min is an integer in subsampling space
				y = y_min;
				l_x = q[low].entry[0];
				r_x = q[middle].entry[0];
			} else {
				y = ceil(y_min);
				l_x = d_l * (y - y_min) + q[low].entry[0];
				r_x = d_r * (y - y_min) + q[middle].entry[0];
			}
			while (y<q[up].entry[1]) { 
				x = ceil(l_x);
				while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_vec = sample_pixel_by_xy_vector(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].vec_fld += new_vec;
						handled[index]++;
					}
					x += 1.0;
				}				
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}
		} else if (q[middle].entry[1] == q[up].entry[1]) { // top horizontal
			if (q[middle].entry[0] > q[up].entry[0]) {
				int temp = middle;
				middle = up;
				up = temp;
			}
			d_l = (q[middle].entry[0]-q[low].entry[0])/(q[middle].entry[1]-q[low].entry[1]);
			d_r = (q[up].entry[0]-q[low].entry[0])/(q[up].entry[1]-q[low].entry[1]);
			
			double y_min = q[low].entry[1];
			if (y_min == ceil(y_min)) {  // special case, y_min is an integer in subsampling space
				y = y_min+1.0;
				l_x = d_l + q[low].entry[0];
				r_x = d_r + q[low].entry[0];
			} else {
				y = ceil(y_min);
				l_x = d_l * (y - y_min) + q[low].entry[0];
				r_x = d_r * (y - y_min) + q[low].entry[0];
			}
			while (y<q[up].entry[1]) { 
				x = ceil(l_x);
				while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_vec = sample_pixel_by_xy_vector(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].vec_fld += new_vec;
						handled[index]++;
					}
					x += 1.0;
				}				
//					y += dy;
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}
		} else {  // no horizontal
			middle_on_right = 1;
			d_l = (q[up].entry[0]-q[low].entry[0])/(q[up].entry[1]-q[low].entry[1]);
			d_r = (q[middle].entry[0]-q[low].entry[0])/(q[middle].entry[1]-q[low].entry[1]);
			if (d_l>d_r) {
				middle_on_right = 0;
				double temp = d_l;
				d_l = d_r;
				d_r = temp;
			}

			// scan lower part of triangle
			double y_min = q[low].entry[1];
			if (y_min == ceil(y_min)) {  // special case, y_min is an integer in subsampling space
				y = y_min+1.0;
				l_x = d_l + q[low].entry[0];
				r_x = d_r + q[low].entry[0];
			} else {
				y = ceil(y_min);
				l_x = d_l * (y - y_min) + q[low].entry[0];
				r_x = d_r * (y - y_min) + q[low].entry[0];
			}
			while (y<q[middle].entry[1]) { 
				x = ceil(l_x);
				while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_vec = sample_pixel_by_xy_vector(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].vec_fld += new_vec;
						handled[index]++;
					}
					x += 1.0;
				}				
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}

			// scan upper half of triangle
			if (middle_on_right == 1) {
				d_r = (q[up].entry[0]-q[middle].entry[0])/(q[up].entry[1]-q[middle].entry[1]);
				r_x = (y - q[middle].entry[1])*d_r + q[middle].entry[0];
			} else {
				d_l = (q[up].entry[0]-q[middle].entry[0])/(q[up].entry[1]-q[middle].entry[1]);
				l_x = (y - q[middle].entry[1])*d_l + q[middle].entry[0];
			}
			while (y<q[up].entry[1]) {
					x = ceil(l_x);
					while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_vec = sample_pixel_by_xy_vector(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].vec_fld += new_vec;
						handled[index]++;
					}
					x += 1.0;
				}				
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}
		}
	}


#if USE_OMP == 1
	#pragma omp for
#endif
	for(i = 0; i < num_pixels; i++) {
		if(handled[i] > 0) 
			fsi.fld[i].vec_fld *= 1.0/handled[i];

		normalize(fsi.fld[i].vec_fld);
	}

#if USE_OMP == 1
	}
#endif


	free(handled);
}


void EdgeFieldSampler::sampleTenFieldToFSI(Polyhedron &poly, FieldSampleImage &fsi) {
	int i;
	int index;
	double l_x, r_x, d_l, d_r;
	double lowest, highest, x, y, t_a;
	int n_horizontal=0;
	int low, middle, up;
	icVector2 q[3];
	int w, h;
	int sw, sh;
	int middle_on_right;
	icMatrix2x2 new_ten;
	unsigned char *handled;
	int num_pixels;


	num_pixels = fsi.w * fsi.h;

	
	sw = fsi.w * num_sample;
	sh = fsi.h * num_sample;

	handled = (unsigned char *)malloc(sizeof(unsigned char) * num_pixels);
	memset(handled, 0, sizeof(unsigned char) * num_pixels);

#if USE_OMP == 1
	#pragma omp parallel private(i, index, l_x, r_x, d_l, d_r, lowest, highest, x, y, t_a, low, middle, up, q, w, h, middle_on_right, new_ten)
	{
#endif


#if USE_OMP == 1
	#pragma omp for
#endif
	for(i = 0; i < num_pixels; i++) {
		fsi.fld[i].ten_fld.set(0.0);
	}


#if USE_OMP == 1
	#pragma omp for
#endif
	for(i = 0; i < poly.ntris; i++) {
		Triangle *temp_t=poly.tlist[i];
		Vertex *v[3];

		for(int j = 0; j < 3; j++) v[j] = temp_t->verts[j];

		if(fabs(v[0]->x) > 0.5 && fabs(v[0]->y) > 0.5 && fabs(v[1]->x) > 0.5 && fabs(v[1]->y) > 0.5 && fabs(v[2]->x) > 0.5 && fabs(v[2]->y) > 0.5)
			continue;

		for(int j = 0; j < 3; j++) q[j].set((v[j]->x + 0.5) * sw, (0.5 - v[j]->y) * sh);


		t_a = poly.calc_triangle_area(q[0], q[1], q[2]);

		low =0;
		up = 0;
		lowest=q[0].entry[1];
		highest=q[0].entry[1];
		if (lowest>q[1].entry[1]) {
			lowest = q[1].entry[1];
			low = 1;
		}
		if (lowest>q[2].entry[1]) {
			lowest = q[2].entry[1];
			low = 2;
		}
		if (highest<q[1].entry[1]) {
			highest = q[1].entry[1];
			up = 1;
		}
		if (highest<q[2].entry[1]) {
			highest = q[2].entry[1];
			up = 2;
		}
		if (low !=0 && up != 0)
			middle =0;
		if (low !=1 && up != 1)
			middle =1;
		if (low !=2 && up != 2)
			middle =2;
	
		if (q[low].entry[1] == q[middle].entry[1]) { // bottom horizontal
			if (q[low].entry[0] > q[middle].entry[0]) {
				int temp = low;
				low = middle;
				middle = temp;
			}
			d_l = (q[up].entry[0]-q[low].entry[0])/(q[up].entry[1]-q[low].entry[1]);
			d_r = (q[up].entry[0]-q[middle].entry[0])/(q[up].entry[1]-q[middle].entry[1]);
			
			double y_min = q[low].entry[1];
			if (y_min == ceil(y_min)) {  // special case, y_min is an integer in subsampling space
				y = y_min;
				l_x = q[low].entry[0];
				r_x = q[middle].entry[0];
			} else {
				y = ceil(y_min);
				l_x = d_l * (y - y_min) + q[low].entry[0];
				r_x = d_r * (y - y_min) + q[middle].entry[0];
			}
			while (y<q[up].entry[1]) { 
				x = ceil(l_x);
				while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_ten = sample_pixel_by_xy_tensor(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].ten_fld += new_ten;
						handled[index]++;
					}
					x += 1.0;
				}				
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}
		} else if (q[middle].entry[1] == q[up].entry[1]) { // top horizontal
			if (q[middle].entry[0] > q[up].entry[0]) {
				int temp = middle;
				middle = up;
				up = temp;
			}
			d_l = (q[middle].entry[0]-q[low].entry[0])/(q[middle].entry[1]-q[low].entry[1]);
			d_r = (q[up].entry[0]-q[low].entry[0])/(q[up].entry[1]-q[low].entry[1]);
			
			double y_min = q[low].entry[1];
			if (y_min == ceil(y_min)) {  // special case, y_min is an integer in subsampling space
				y = y_min+1.0;
				l_x = d_l + q[low].entry[0];
				r_x = d_r + q[low].entry[0];
			} else {
				y = ceil(y_min);
				l_x = d_l * (y - y_min) + q[low].entry[0];
				r_x = d_r * (y - y_min) + q[low].entry[0];
			}
			while (y<q[up].entry[1]) { 
				x = ceil(l_x);
				while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_ten = sample_pixel_by_xy_tensor(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].ten_fld += new_ten;
						handled[index]++;
					}
					x += 1.0;
				}				
//					y += dy;
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}
		} else {  // no horizontal
			middle_on_right = 1;
			d_l = (q[up].entry[0]-q[low].entry[0])/(q[up].entry[1]-q[low].entry[1]);
			d_r = (q[middle].entry[0]-q[low].entry[0])/(q[middle].entry[1]-q[low].entry[1]);
			if (d_l>d_r) {
				middle_on_right = 0;
				double temp = d_l;
				d_l = d_r;
				d_r = temp;
			}

			// scan lower part of triangle
			double y_min = q[low].entry[1];
			if (y_min == ceil(y_min)) {  // special case, y_min is an integer in subsampling space
				y = y_min+1.0;
				l_x = d_l + q[low].entry[0];
				r_x = d_r + q[low].entry[0];
			} else {
				y = ceil(y_min);
				l_x = d_l * (y - y_min) + q[low].entry[0];
				r_x = d_r * (y - y_min) + q[low].entry[0];
			}
			while (y<q[middle].entry[1]) { 
				x = ceil(l_x);
				while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_ten = sample_pixel_by_xy_tensor(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].ten_fld += new_ten;
						handled[index]++;
					}
					x += 1.0;
				}				
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}

			// scan upper half of triangle
			if (middle_on_right == 1) {
				d_r = (q[up].entry[0]-q[middle].entry[0])/(q[up].entry[1]-q[middle].entry[1]);
				r_x = (y - q[middle].entry[1])*d_r + q[middle].entry[0];
			} else {
				d_l = (q[up].entry[0]-q[middle].entry[0])/(q[up].entry[1]-q[middle].entry[1]);
				l_x = (y - q[middle].entry[1])*d_l + q[middle].entry[0];
			}
			while (y<q[up].entry[1]) {
					x = ceil(l_x);
					while (x < r_x) {  // doesn't matter whether r_x is an integer or not in the subsample domain
					new_ten = sample_pixel_by_xy_tensor(poly, x, y, q, temp_t, t_a, &w, &h);
					if ((w>=0 && w<fsi.w) && (h>=0 && h<fsi.h)) {
						index = w + h * fsi.w;
						fsi.fld[index].ten_fld += new_ten;
						handled[index]++;
					}
					x += 1.0;
				}				
				y += 1.0;
				l_x += d_l;
				r_x += d_r;
			}
		}
	}


#if USE_OMP == 1
	#pragma omp for
#endif
	for(i = 0; i < num_pixels; i++) {
		if (handled[i] > 0) 
			fsi.fld[i].ten_fld *= 1.0/handled[i];

		if (fsi.fld[i].ten_fld.entry[0][0] == 0.0 && fsi.fld[i].ten_fld.entry[0][1] == 0.0) {
			fsi.fld[i].vec_fld.set(0.0);
		}
		else {
			double temp_angle = atan2(fsi.fld[i].ten_fld.entry[0][1], fsi.fld[i].ten_fld.entry[0][0])*0.5;
			fsi.fld[i].vec_fld.set(cos(temp_angle), sin(temp_angle));
		}
	}

#if USE_OMP == 1
	}
#endif

	free(handled);
}


void EdgeFieldSampler::sampleTenFieldRegionToMesh(Polyhedron & poly, FieldSampleImage& fsi,  int* RegionIDArray) {
	int width, height, num_verts;
	int wx, wy, wx_l, wy_u;
	double distance,minDistance;
	int pixelX, pixelY;
	Vertex *v;
	

	num_verts = poly.nverts;
	width = fsi.w;
	height = fsi.h;
	/////////////////////////////////////////////////////////////////////////////////////////////////
	//output
    /////////////////////////////////////////////////////////////////////////////////////////////////

	/*output++;
	CString num;
	num.Format(_T("vertex_region%d.txt"),output);
	ofstream outfile(num);*/
	///////////////////////////////////////////////////////////////////////

	for(int i = 0; i < num_verts; i++) {
		v = poly.vlist[i];
		
		wx = (int)((v->x + 0.5) * width);
		wy = (int)((0.5 - v->y) * height);
		wx_l = wx + 1;
		wy_u = wy + 1;
		
		//initial (-1 is reserved for global region)
		v->regionID = -2;
		
		minDistance=10000000.0;
		
		pixelX=-1;
		pixelY=-1;

		if(wx >= 0 && wx < width) {
			if(wy >= 0 && wy < height) {
				//compute the distance
				distance=sqrt((wx-v->x* width)*(wx-v->x* width)+(wy-v->y* height)*(wy-v->y* height));
				//get the minDistance
				if(distance<=minDistance)
				{
					minDistance=distance;
					pixelX=wx;
		            pixelY=wy;
				}
			}

			if(wy_u >= 0 && wy_u < height) {
				//compute the distance
				distance=sqrt((wx-v->x* width)*(wx-v->x* width)+(wy_u-v->y* height)*(wy_u-v->y* height));
				//get the minDistance
				if(distance<=minDistance)
				{
					minDistance=distance;
					pixelX=wx;
		            pixelY=wy_u;
				}
			}
		}

		if(wx_l >= 0 && wx_l < width) {
			if(wy >= 0 && wy < height) {
				//compute the distance
				distance=sqrt((wx_l-v->x* width)*(wx_l-v->x* width)+(wy-v->y* height)*(wy-v->y* height));
				//get the minDistance
				if(distance<=minDistance)
				{
					minDistance=distance;
					pixelX=wx_l;
		            pixelY=wy;
				}
			}

			if(wy_u >= 0 && wy_u < height) {
				//compute the distance
				distance=sqrt((wx_l-v->x* width)*(wx_l-v->x* width)+(wy-v->y* height)*(wy-v->y* height));
				//get the minDistance
				if(distance<=minDistance)
				{
					minDistance=distance;
					pixelX=wx_l;
		            pixelY=wy_u;
				}
			}
		}

        if(pixelX>=0&&pixelY>=0)
			v->regionID=RegionIDArray[pixelY * width + pixelX];

		poly.vlist[i]->regionID=v->regionID;
		//outfile<<"( "<<v->x<<","<<v->y<<","<<pixelX<<","<<pixelY<<","<<v->regionID<<"), "<<endl;

		
	}
	/*for(int i = 0; i < num_verts; i++)
		outfile<<"( "<<poly.vlist[i]->x<<","<<poly.vlist[i]->y<<","<<poly.vlist[i]->regionID<<"), "<<endl;

	outfile.close();*/
}
