#pragma once
#include "tendes_plane.h"
#include "image_pyramid.h"
#include "icVector.h"
#include "icMatrix.h"
#include "math.h"
#include <fstream>
#include <iostream>
#include <sstream>


/*
	EdgeFieldSampler:

	Sample vector/tensor field from FSI to Polyhedron or do the samping backward.
	
	It assumes that the coordinate system of FSI is such that the origin is at top-left and its y-axis runs from top to bottom,
	whereas the origin of Polyhedron lies at center, and the y-axis goes upward. 

	If any vertex in Polyhedron lies in x:[-.5, .5], y:[-.5, 5], then it will be sampled; ignored otherwise.
*/
class EdgeFieldSampler {
private:
	static const int num_sample = 1;


	void sampleVecFieldToFSI(Polyhedron &poly, FieldSampleImage &fsi);
	void sampleTenFieldToFSI(Polyhedron &poly, FieldSampleImage &fsi);

	void sampleVecFieldToMesh(Polyhedron &poly, FieldSampleImage &fsi);
	void sampleTenFieldToMesh(Polyhedron &poly, FieldSampleImage &fsi);

	inline icVector2 sample_pixel_by_xy_vector(Polyhedron &poly, double x, double y, icVector2 q[3], Triangle *temp_t, double t_a, int *w, int *h);
	inline icMatrix2x2 sample_pixel_by_xy_tensor(Polyhedron &poly, double x, double y, icVector2 q[3], Triangle *temp_t, double t_a, int *w, int *h);

public:
	void sampleToMesh(Polyhedron &poly, FieldSampleImage &fsi);
	void sampleToFSI(Polyhedron &poly, FieldSampleImage &fsi);
	void sampleTenFieldRegionToMesh(Polyhedron & poly, FieldSampleImage& fsi,  int* RegionIDArray);
};