#pragma once
#include "IPaintRenderer.h"


class ILayeredStrokePaintRenderer : public IPaintRenderer {
protected:
	int max_layers;
	int render_mode;
	int seed_gen_mode;

public:
	// create a multi-layer renderer using strokes
	ILayeredStrokePaintRenderer(int width, int height, int num_layers) : IPaintRenderer(width, height) { 
		if(num_layers <= 0) {
			throw "ILayeredStrokePaintRenderer: Invalid Layers\r\n";
		}

		max_layers = num_layers; 
		render_mode = 0;
		seed_gen_mode = 0;
	}

	virtual ~ILayeredStrokePaintRenderer() {}

	// change the number of layers
	virtual bool setNumLayers(int num_layers) = 0;

	// return the number of supported
	int getNumLayers() { return max_layers; }

	// move each seed according to 2D optical flow. The parameter is assumed to have the same dimension as this renderer
	virtual void pushSeeds(int layer, const float *opt_flow, int width, int height) = 0;

	// clear seeds at a specified layer
	virtual void clearSeeds(int layer) = 0;

	// return the number of strokes
	virtual size_t getNumStrokes(int layer) = 0;

	// get rendering time
	virtual double getRenderingTime() = 0;

	// 0: normal, 1: height-map... could be extensible
	virtual void setRenderMode(int mode) {
		if(mode == 0 || mode == 1) render_mode = mode;
	}

	// 0: regenerate, 1: update
	virtual void setGenMode(int mode) {
		if(mode == 0 || mode == 1) seed_gen_mode = mode;
	}
};
