#pragma once
#include "stdafx.h"
#include "ImageLoader.h"
#include <string>
#include <math.h>
#include "corona.h"
#include "ppm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#define   malloc(s)         _malloc_dbg(s, _NORMAL_BLOCK, __FILE__, __LINE__)
#endif


using namespace std;
using namespace corona;


// Constructor/Destructor
ImageLoader::ImageLoader() {
	this->padding_color[0] = 0;
	this->padding_color[1] = 0;
	this->padding_color[2] = 0;
}


ImageLoader::~ImageLoader() {
}


// Public member functions
void ImageLoader::load(const char *filename, bool fill_padding, unsigned char **pixels, int &w, int &h) {
	this->readFile(string(filename), fill_padding, pixels, w, h);
}


void ImageLoader::load(string filename, bool fill_padding, unsigned char **pixels, int &w, int &h) {
	this->readFile(filename, fill_padding, pixels, w, h);
}


void ImageLoader::addPaddingBottom(int num_pixels, unsigned char **pixels, int &w, int &h) {
	unsigned char *new_pixels;
	int new_width, new_height;
	int idx, idx2;

	new_width = w;
	new_height = h + num_pixels;
	new_pixels = new unsigned char[3 * new_width * new_height];


	for(int i = 0; i < num_pixels; i++) {
		for(int j = 0; j < w; j++) {
			idx = j + i * w;

			new_pixels[3 * idx + 0] = this->padding_color[0];
			new_pixels[3 * idx + 1] = this->padding_color[1];
			new_pixels[3 * idx + 2] = this->padding_color[2];
		}
	}

	for(int i = 0; i < h; i++) {
		for(int j = 0; j < w; j++) {
			idx = j + (num_pixels + i) * w;
			idx2 = j + i * w;

			new_pixels[3 * idx + 0] = (*pixels)[3 * idx2 + 0];
			new_pixels[3 * idx + 1] = (*pixels)[3 * idx2 + 1];
			new_pixels[3 * idx + 2] = (*pixels)[3 * idx2 + 2];
		}
	}

	w = new_width;
	h = new_height;

	delete[] *pixels;
	*pixels = new_pixels;
}


void ImageLoader::addPaddingRight(int num_pixels, unsigned char **pixels, int &w, int &h) {
	unsigned char *new_pixels;
	int new_width, new_height;
	int idx, idx2;

	new_width = w + num_pixels;
	new_height = h;
	new_pixels = new unsigned char[3 * new_width * new_height];


	for(int i = 0; i < h; i++) {
		for(int j = 0; j < w; j++) {
			idx = j + i * new_width;
			idx2 = j + i * w;

			new_pixels[3 * idx + 0] = (*pixels)[3 * idx2 + 0];
			new_pixels[3 * idx + 1] = (*pixels)[3 * idx2 + 1];
			new_pixels[3 * idx + 2] = (*pixels)[3 * idx2 + 2];
		}

		for(int j = 0; j < num_pixels; j++) {
			idx = (j + w) + i * new_width;
			idx2 = (i + 1) * w - 1;

			new_pixels[3 * idx + 0] = (*pixels)[3 * idx2 + 0];
			new_pixels[3 * idx + 1] = (*pixels)[3 * idx2 + 1];
			new_pixels[3 * idx + 2] = (*pixels)[3 * idx2 + 2];
		}
	}

	w = new_width;
	h = new_height;

	delete[] *pixels;
	*pixels = new_pixels;
}


void ImageLoader::setPaddingColor(const unsigned char rgb[3]) {
	if(rgb != NULL) {
		for(int i = 0; i < 3; i++) {
			this->padding_color[i] = rgb[i];
		}
	}
}


void ImageLoader::getPaddingColor(unsigned char rgb[3]) {
	if(rgb != NULL) {
		for(int i = 0; i < 3; i++) {
			rgb[i] = this->padding_color[i];
		}
	}
}


void ImageLoader::squarize(unsigned char **pixels, int &w, int &h) {
	int nw, nh;

	nw = 1;
	nh = 1;

	while(nw < w) nw *= 2;
	while(nh < h) nh *= 2;

	if(nw > nh) nh = nw;
	else nw = nh;

	this->addPaddingRight(nw - w, pixels, w, h);
	this->addPaddingBottom(nh - h, pixels, w, h);

	w = nw;
	h = nh;
}


void ImageLoader::readFile(string filename, bool fill_padding, unsigned char **pixels, int &w, int &h) {
	string fn_cp = filename;

	for(int i = 0; i < (int)fn_cp.length(); i++) fn_cp[i] = tolower(fn_cp[i]);

	if(fn_cp.rfind(".ppm") != string::npos) {
		this->readPPM(filename.c_str(), pixels, w, h);
	}
	else if(fn_cp.rfind(".jpeg") != string::npos || fn_cp.rfind(".jpg") != string::npos) {
		this->readJPG(filename.c_str(), pixels, w, h);
	}
	else if(fn_cp.rfind(".bmp") != string::npos) {
		this->readBMP(filename.c_str(), pixels, w, h);
	}

	if(w > 0 && h > 0 && fill_padding) {
		this->squarize(pixels, w, h);
	}
}


void ImageLoader::readPPM(const char *filename, unsigned char **pixels, int &w, int &h) {
	::My_SSS_ReadPPM(&w, &h, pixels, filename);
}



void ImageLoader::readJPG(const char *filename, unsigned char **pixels, int &w, int &h) {
	unsigned char *c_ptr = NULL;
	corona::Image *jpgImg = NULL;

	jpgImg = corona::OpenImage(filename, FF_JPEG, PF_R8G8B8);

	if(jpgImg) {
		w = jpgImg->getWidth();
		h = jpgImg->getHeight();

		c_ptr = (unsigned char *)jpgImg->getPixels();
		*pixels = new unsigned char[3 * w * h];

		//for(int i = h - 1; i >= 0; i--) {
		for(int i = 0; i < h; i++) {
			for(int j = 0; j < w; j++) {
				(*pixels)[3 * (j + i * w)] = *c_ptr++;
				(*pixels)[3 * (j + i * w) + 1] = *c_ptr++;
				(*pixels)[3 * (j + i * w) + 2] = *c_ptr++;
			}
		}

		delete jpgImg;
	}
	else {
		w = -1;
		h = -1;
	}
}


void ImageLoader::readBMP(const char *filename, unsigned char **pixels, int &w, int &h) {
	unsigned char *c_ptr = NULL;
	corona::Image *bmpImg = NULL;

	bmpImg = corona::OpenImage(filename, FF_BMP, PF_R8G8B8);

	if(bmpImg) {
		w = bmpImg->getWidth();
		h = bmpImg->getHeight();

		c_ptr = (unsigned char *)bmpImg->getPixels();
		*pixels = new unsigned char[3 * w * h];

		//for(int i = h - 1; i >= 0; i--) {
		for(int i = 0; i < h; i++) {
			for(int j = 0; j < w; j++) {
				(*pixels)[3 * (j + i * w)] = *c_ptr++;
				(*pixels)[3 * (j + i * w) + 1] = *c_ptr++;
				(*pixels)[3 * (j + i * w) + 2] = *c_ptr++;
			}
		}

		delete bmpImg;
	}
	else {
		w = -1;
		h = -1;
	}
}

