#pragma once
#include "StdAfx.h"
#include "PaintingPreset.h"
#include "PaintingStyleField.h"
#include "Util.h"



inline float blend(float v1, float v2, float w1, float w2) { return v1 * w1 + v2 * w2; }

inline float linear(float t) { return t; }
inline float square1(float t) { return 1 - sqrt(1.f - t * t); }
inline float square2(float t) { float omt = 1.f - t; return sqrt(1.f - omt * omt); }
inline float cube1(float t) { return 1 - pow(1.f - t * t * t, 1.f/3.f); }
inline float cube2(float t) { float omt = 1.f - t; return pow(1.f - omt * omt * omt, 1.f/3.f); }
inline float ease(float t) { return (sin((t - .5f) * M_PI) + 1.f) * .5f; }


void blendParams(const PaintingPreset &p1, const PaintingPreset &p2, float w1, float w2, PaintingPreset &result) {
	result.num_layers				= (int)(blend(p1.num_layers, p2.num_layers, w1, w2) + .5);
	result.displacement				= blend(p1.displacement, p2.displacement, w1, w2);
	result.area_error				= blend(p1.area_error, p2.area_error, w1, w2);


	// jitter setting
	result.jitter_amplitude			= blend(p1.jitter_amplitude, p2.jitter_amplitude, w1, w2);


	// stroke setting
	result.stroke_initial_opacity	= blend(p1.stroke_initial_opacity, p2.stroke_initial_opacity, w1, w2);
	result.stroke_pos_jitter		= blend(p1.stroke_pos_jitter, p2.stroke_pos_jitter, w1, w2);
	result.stroke_size				= blend(p1.stroke_size, p2.stroke_size, w1, w2);
	result.stroke_gridsize			= blend(p1.stroke_gridsize, p2.stroke_gridsize, w1, w2);

	result.stroke_min_length		= (int)(blend(p1.stroke_min_length, p2.stroke_min_length, w1, w2) + .5);
	result.stroke_max_length		= (int)(blend(p1.stroke_max_length, p2.stroke_max_length, w1, w2) + .5);

	result.stroke_color_shift[0]	= blend(p1.stroke_color_shift[0], p2.stroke_color_shift[0], w1, w2);
	result.stroke_color_shift[1]	= blend(p1.stroke_color_shift[1], p2.stroke_color_shift[1], w1, w2);
	result.stroke_color_shift[2]	= blend(p1.stroke_color_shift[2], p2.stroke_color_shift[2], w1, w2);


	// post processing setting
	result.post_process_val1		= blend(p1.post_process_val1, p2.post_process_val1, w1, w2);
	result.post_process_val2		= blend(p1.post_process_val2, p2.post_process_val2, w1, w2);
	result.post_process_val3		= blend(p1.post_process_val3, p2.post_process_val3, w1, w2);
	result.post_process_val4		= blend(p1.post_process_val4, p2.post_process_val4, w1, w2);
}


void interpolatePaintingPreset(const PaintingPreset &p1, const PaintingPreset &p2, float param, PaintingPreset &result, int func_type) {
	float weight = 0;


	// fix these values...
	result.alpha = PaintingStyleField::mp_alpha;
	result.post_process_operation = PaintingStyleField::mp_postprocessing_type;
	result.color_composition_method = PaintingStyleField::mp_color_composition;

	param = (float)::clampVal(param, 0, 1.);

	switch(func_type) {
		case 0:
		default:
			weight = linear(param);
			break;

		case 1:
			weight = square1(param);
			break;

		case 2:
			weight = cube1(param);
			break;

		case 3:
			weight = square2(param);
			break;

		case 4:
			weight = cube2(param);
			break;

		case 5:
			weight = ease(param);
			break;
	}

	blendParams(p1, p2, 1.f - weight, weight, result);
}

