#pragma once
#include "StdAfx.h"
#include "PathMaker.h"



PathMaker::PathMaker() {
	x0 = y0 = 0;
	mode = FREEFORM;
	points = NULL;
}


PathMaker::~PathMaker() {
}


PathMaker::MODE PathMaker::getMode() {
	return mode;
}


void PathMaker::setMode(PathMaker::MODE new_mode) {
	mode = new_mode;
}


void PathMaker::start(int x0, int y0) {

	if(points == NULL) return;

	points->clear();
	this->x0 = x0;
	this->y0 = y0;

	switch(mode) {
		case FREEFORM:
		case LINE:
			points->push_back(pair<int, int>(x0, y0));
			break;

		default:
			break;
	}
}


void PathMaker::update(int x, int y) {
	int w, h;
	int num_pts = 36;
	double r, theta, offset, angle;

	if(points == NULL) return;

	switch(mode) {
		case FREEFORM:
			points->push_back(pair<int, int>(x, y));
			break;

		case CIRCLE:
			w = x - x0;
			h = y - y0;
			r = sqrt((double)(w * w + h * h));
			theta = 2 * M_PI / num_pts;
			offset = M_PI / 2;

			points->clear();

			for(int i = 0; i < num_pts; i++) {
				angle = offset + i * theta;
				points->push_back(pair<int, int>((int)(x0 + r * cos(angle)), (int)(y0 + r * sin(angle))));
			}

			points->push_back(pair<int, int>(points->at(0).first, points->at(0).second));

			break;

		case SQUARE:
			w = x - x0;
			h = y - y0;
			points->clear();
			points->push_back(pair<int, int>(x0, y0));
			points->push_back(pair<int, int>(x0 + w, y0));
			points->push_back(pair<int, int>(x0 + w, y0 + h));
			points->push_back(pair<int, int>(x0, y0 + h));
			points->push_back(pair<int, int>(x0, y0));

			break;


		case LINE:
			if(points->size() <= 1) {
				points->push_back(pair<int, int>(x, y));
			}
			else {
				points->pop_back();
				points->push_back(pair<int, int>(x, y));
			}

			break;

		default:
			break;
	}
}


void PathMaker::getCenter(pair<int, int> &center) {
	switch(mode) {
		case FREEFORM:
			if(points->size() > 0) {
				int x = 0, y = 0;

				for(size_t i = 0; i < points->size(); i++) {
					x += points->at(i).first;
					y += points->at(i).second;
				}

				center.first = (int)(x / points->size());
				center.second = (int)(y / points->size());
			}
			else {
				center.first = x0;
				center.second = y0;
			}

			break;

		case CIRCLE:
			center.first = x0;
			center.second = y0;
			break;

		case SQUARE:
			if(points->size() == 5) {
				center.first = (points->at(0).first + points->at(1).first) / 2;
				center.second = (points->at(0).second + points->at(2).second) / 2;
			}
			else {
				center.first = x0;
				center.second = y0;
			}

			break;

		case LINE:
			if(points->size() >= 2) {
				center.first = (points->at(0).first + points->at(1).first) / 2;
				center.second = (points->at(0).second + points->at(1).second) / 2;
			}
			else {
				center.first = x0;
				center.second = y0;
			}

		default:
			break;
	}
}


void PathMaker::setPathContainer(vector< pair<int, int> >  &container) {
	points = &container;
}


void PathMaker::releasePathContainer() {
	points = NULL;
	x0 = 0;
	y0 = 0;
}
