uniform sampler2D tex0;
uniform sampler2D tex1;
uniform float hue_breakpoint;		// must be in [0, 360.0)
uniform int h_func;
uniform int s_func;
uniform int v_func;
uniform int eval;
uniform float bgfg_threshold;


//
//  0: HSV, 
//  1: SVH, 
//  2: VHS, 
//  3: VSH, 
//  4: HVS, 
//  5: SHV, 
//


vec3 RgbHsv(in vec4 RGB) {
	float ch_min = min(RGB.r, min(RGB.g, RGB.b));
	float ch_max = max(RGB.r, max(RGB.g, RGB.b));
	float delta = ch_max - ch_min;
	vec3 HSV = vec3(0.0, 0.0, ch_max);


	if(ch_max != 0.0) {
		HSV[1] = delta / ch_max;

		if(RGB[0] == ch_max) {
			HSV[0] = (RGB[1] - RGB[2]) / delta;
		}
		else if(RGB[1] == ch_max) {
			HSV[0] = 2.0 + (RGB[2] - RGB[0]) / delta;
		}
		else {
			HSV[0] = 4.0 + (RGB[0] - RGB[1]) / delta;
		}

		HSV[0] *= 60.0;

		if(HSV[0] < 0.0) {
			HSV[0] += 360.0;
		}
		
		if(HSV[0] >= 360.0) {
			HSV[0] -= 360.0;
		}
	}
	else {
		HSV[1] = 0.0;
		HSV[0] = 0.0;
	}

	return HSV;
}


void clamp1(inout float hue) {
	hue -= hue_breakpoint;

	if(hue < 0.0) hue += 360.0;
}


void clamp2(inout float hue, inout bool flipped) {
	clamp1(hue);

	if(hue > 180.0) {
		flipped = true;
		hue = 360.0 - hue;
	}
	else {
		flipped = false;
	}
}

int isMin_H(in float h0, in float h1) {
	clamp1(h0);
	clamp1(h1);

	return (h0 < h1) ? 1 : ((h0 > h1) ? -1 : 0);
}


int isMax_H(in float h0, in float h1) {
	clamp1(h0);
	clamp1(h1);

	return (h0 > h1) ? 1 : ((h0 < h1) ? -1 : 0);
}


int isWarm1_H(in float h0, in float h1) {
	bool flip[2];

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 < h1) return 1;
	else if(h1 < h0) return -1;
	else if(flip[0] == false && flip[1] == true) return 1;
	else if(flip[0] == true && flip[1] == false) return -1;
	else return 0;
}


int isCool1_H(in float h0, in float h1) {
	bool flip[2];

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 > h1) return 1;
	else if(h1 > h0) return -1;
	else if(flip[0] == false && flip[1] == true) return 1;
	else if(flip[0] == true && flip[1] == false) return -1;
	else return 0;
}


int isWarm2_H(in float h0, in float h1) {
	bool flip[2];

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 < h1) return 1;
	else if(h1 < h0) return -1;
	else if(flip[0] == true && flip[1] == false) return 1;
	else if(flip[0] == false && flip[1] == true) return -1;
	else return 0;
}


int isCool2_H(in float h0, in float h1) {
	bool flip[2];

	clamp2(h0, flip[0]);
	clamp2(h1, flip[1]);

	if(h0 > h1) return 1;
	else if(h1 > h0) return -1;
	else if(flip[0] == true && flip[1] == false) return 1;
	else if(flip[0] == false && flip[1] == true) return -1;
	else return 0;
}


int isMin_S(in float s0, in float s1) { return (s0 < s1) ? 1 : ((s0 > s1) ? -1 : 0); }
int isMax_S(in float s0, in float s1) { return (s0 > s1) ? 1 : ((s0 < s1) ? -1 : 0); }
int isMin_V(in float v0, in float v1) { return (v0 < v1) ? 1 : ((v0 > v1) ? -1 : 0); }
int isMax_V(in float v0, in float v1) { return (v0 > v1) ? 1 : ((v0 < v1) ? -1 : 0); }


int Hue_Func(in float h0, in float h1) {
	if(h_func == 0) {					// min.
		return isMin_H(h0, h1);
	}
	else if(h_func == 1) {				// max.
		return isMax_H(h0, h1);
	}
	else if(h_func == 2) {				// warm 1
		return isWarm1_H(h0, h1);
	}
	else if(h_func == 3) {				// cool 1
		return isCool1_H(h0, h1);
	}
	else if(h_func == 4) {				// warm 2
		return isWarm2_H(h0, h1);
	}
	else if(h_func == 5) {				// cool 2
		return isCool2_H(h0, h1);
	}
	else {
		return -1;
	}
}


int Sat_Func(in float s0, in float s1) {
	if(s_func == 0) {					// min.
		return isMin_S(s0, s1);
	}
	else if(s_func == 1) {				// max.
		return isMax_S(s0, s1);
	}
	else {
		return -1;
	}
}


int Val_Func(in float v0, in float v1) {
	if(v_func == 0) {					// min.
		return isMin_V(v0, v1);
	}
	else if(v_func == 1) {				// max.
		return isMax_V(v0, v1);
	}
	else {
		return -1;
	}
}



void main() {
	vec4 c0 = texture2D(tex0, gl_TexCoord[0].st);
	vec4 c1 = texture2D(tex1, gl_TexCoord[0].st);
	vec3 hsv0 = RgbHsv(c0);
	vec3 hsv1 = RgbHsv(c1);
	int result[2];


	if(c0.a < bgfg_threshold) {
		gl_FragColor = c1;
		return;
	}
	else if(c1.a < bgfg_threshold) {
		gl_FragColor = c0;
		return;
	}

	if(eval == 0) {		// HSV
		result[0] = Hue_Func(hsv0[0], hsv1[0]);

		if(result[0] == 1) {
			gl_FragColor = c0;
		}
		else if(result[0] == 0) {
			result[1] = Sat_Func(hsv0[1], hsv1[1]);

			if(result[1] == 1) {
				gl_FragColor = c0;
			}
			else if(result[1] == 0) {
				if(Val_Func(hsv0[2], hsv1[2]) == 1) {
					gl_FragColor = c0;
				}
				else {
					gl_FragColor = c1;
				}
			}
			else {
				gl_FragColor = c1;
			}
		}
		else {
			gl_FragColor = c1;
		}
	}
	else if(eval == 1) {		// SVH
		result[0] = Sat_Func(hsv0[1], hsv1[1]);

		if(result[0] == 1) {
			gl_FragColor = c0;
		}
		else if(result[0] == 0) {
			result[1] = Val_Func(hsv0[2], hsv1[2]);

			if(result[1] == 1) {
				gl_FragColor = c0;
			}
			else if(result[1] == 0) {
				if(Hue_Func(hsv0[0], hsv1[0]) == 1) {
					gl_FragColor = c0;
				}
				else {
					gl_FragColor = c1;
				}
			}
			else {
				gl_FragColor = c1;
			}
		}
		else {
			gl_FragColor = c1;
		}
	}
	else if(eval == 2) {		// VHS
		result[0] = Val_Func(hsv0[2], hsv1[2]);

		if(result[0] == 1) {
			gl_FragColor = c0;
		}
		else if(result[0] == 0) {
			result[1] = Hue_Func(hsv0[0], hsv1[0]);

			if(result[1] == 1) {
				gl_FragColor = c0;
			}
			else if(result[1] == 0) {
				if(Sat_Func(hsv0[1], hsv1[1]) == 1) {
					gl_FragColor = c0;
				}
				else {
					gl_FragColor = c1;
				}
			}
			else {
				gl_FragColor = c1;
			}
		}
		else {
			gl_FragColor = c1;
		}
	}
	else if(eval == 3) {		// VSH
		result[0] = Val_Func(hsv0[2], hsv1[2]);

		if(result[0] == 1) {
			gl_FragColor = c0;
		}
		else if(result[0] == 0) {
			result[1] = Sat_Func(hsv0[1], hsv1[1]);

			if(result[1] == 1) {
				gl_FragColor = c0;
			}
			else if(result[1] == 0) {
				if(Hue_Func(hsv0[0], hsv1[0]) == 1) {
					gl_FragColor = c0;
				}
				else {
					gl_FragColor = c1;
				}
			}
			else {
				gl_FragColor = c1;
			}
		}
		else {
			gl_FragColor = c1;
		}
	}
	else if(eval == 4) {		// HVS
		result[0] = Hue_Func(hsv0[0], hsv1[0]);

		if(result[0] == 1) {
			gl_FragColor = c0;
		}
		else if(result[0] == 0) {
			result[1] = Val_Func(hsv0[2], hsv1[2]);

			if(result[1] == 1) {
				gl_FragColor = c0;
			}
			else if(result[1] == 0) {
				if(Sat_Func(hsv0[1], hsv1[1]) == 1) {
					gl_FragColor = c0;
				}
				else {
					gl_FragColor = c1;
				}
			}
			else {
				gl_FragColor = c1;
			}
		}
		else {
			gl_FragColor = c1;
		}
	}
	else if(eval == 5) {		// SHV
		result[0] = Sat_Func(hsv0[1], hsv1[1]);

		if(result[0] == 1) {
			gl_FragColor = c0;
		}
		else if(result[0] == 0) {
			result[1] = Hue_Func(hsv0[0], hsv1[0]);

			if(result[1] == 1) {
				gl_FragColor = c0;
			}
			else if(result[1] == 0) {
				if(Val_Func(hsv0[2], hsv1[2]) == 1) {
					gl_FragColor = c0;
				}
				else {
					gl_FragColor = c1;
				}
			}
			else {
				gl_FragColor = c1;
			}
		}
		else {
			gl_FragColor = c1;
		}
	}
	else {
		gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);
	}
}


