#include "stdafx.h"
#include "SplineRenderer.h"
#include <algorithm>
#include <time.h>
#include "cv.h"
#include "cvaux.h"
#include "StrokeMapManager.h"
#include "shader_flag.h"
#include "renderer_debug.h"
#include "ppm.h"
#include "VisView.h"


#if D_SAVESTAT_SPLINE == 1
#include "debug_timing.h"
#else 
#include "debug_no_timing.h"
#endif


int d_spline_num_steps = 0, d_spline_num_grows = 0;
int d_spline_num_refimgcomps = 0, d_spline_num_cvblur_cvcopy = 0, d_spline_num_clears = 0, d_spline_num_listcomps = 0;
__int64 **d_spline_step = NULL, **d_spline_grows = NULL;
__int64 **d_spline_refimgcomp = NULL, **d_spline_cvblur_cvcopy = NULL, **d_spline_clear = NULL, **d_spline_listcomp = NULL;




typedef list<SimpleStroke*> seed_list;
typedef list<GLSpline*> spline_list;


/*
		ref:		0 ~ 255
		noise:		0 ~ 1.f
*/
inline void Spline_sampleColor(unsigned char color[3], unsigned char ref[3], float noise[3], float amp) {
	color[0] = (unsigned char)::clampVal((ref[0] + noise[0] * amp), 0, 255);
	color[1] = (unsigned char)::clampVal((ref[1] + noise[1] * amp), 0, 255);
	color[2] = (unsigned char)::clampVal((ref[2] + noise[2] * amp), 0, 255);
}


/*
		ref:		0 ~ 255
		shift:		0 ~ 255.f
		noise:		0 ~ 1.f
*/
inline void Spline_sampleColorJitter(unsigned char color[3], unsigned char ref[3], float shift[3], float noise[3], float amp) {
	/**
	color[0] = (unsigned char)::clampVal((ref[0] + noise[0] * amp), 0, 255);
	color[1] = (unsigned char)::clampVal((ref[1] + noise[1] * amp), 0, 255);
	color[2] = (unsigned char)::clampVal((ref[2] + noise[2] * amp), 0, 255);
	/**/

	/**/
	float hsv[3];
	RgbHsv(ref, hsv);

	// never jitter color when a color if fully-desaturated
	if(hsv[1] > 0 && hsv[2] > 0) {
		hsv[0] = hsv[0] + shift[0];
		hsv[1] = (float)::clampVal((hsv[1] + shift[1]), 0, 1.f);
		hsv[2] = (float)::clampVal((hsv[2] + shift[2]), 0, 1.f);
	}

	HsvRgb(hsv, color);
	color[0] = (unsigned char)::clampVal((color[0] + noise[0] * amp), 0, 255);
	color[1] = (unsigned char)::clampVal((color[1] + noise[1] * amp), 0, 255);
	color[2] = (unsigned char)::clampVal((color[2] + noise[2] * amp), 0, 255);
	/**/
}


// assume the triplet is in RGB order.
inline float calcDepth(unsigned char color[3]) {
	__int64 max_depth = 0x000002fa03ffffff;
	__int64 unscaled = ((__int64)(color[0]*color[0]+color[1]*color[1]+color[2]*color[2]) << 24) + ((__int64)color[1] << 16) + ((__int64)color[0] << 8) + (__int64)color[2];
	return unscaled / (float)max_depth;
}


Spline_SimpleLayer::Spline_SimpleLayer(int width, int height) {
	rendered_image = new unsigned char[4 * width * height];	
	initial_stroke_image = new unsigned char[4 * width * height];
	strokes.resize(0);
	splines.resize(0);
}


Spline_SimpleLayer::~Spline_SimpleLayer() {
	delete[] rendered_image;
	delete[] initial_stroke_image;

	while(strokes.empty() == false) {
		delete strokes.front();
		strokes.pop_front();
	}

	while(splines.empty() == false) {
		delete splines.front();
		splines.pop_front();
	}
}




const float SplineRenderer::quad_vertex_coords[] = {
	-1.f, -1.f, 
	 1.f, -1.f, 
	 1.f,  1.f, 
	-1.f,  1.f, 
};

const float SplineRenderer::quad_texture_coords[] = {
	0.0, 0.0,
	1.f, 0.0,
	1.f, 1.f, 
	0.0, 1.f, 
};




SplineRenderer::SplineRenderer(int width, int height, int num_layers) : ILayeredStrokePaintRenderer(width, height, num_layers) {
	initializeBuffer();
	makeStrokeTemplate();
}


SplineRenderer::~SplineRenderer() {
	finalizeBuffer();
	cleanSeedTemplate();
}


void SplineRenderer::initializeBuffer() {
	int num_elems;
	const float noise_gen_factor = 2.f / (float)RAND_MAX;

	srand(0);

	num_pixels = width * height;
	num_elems = 3 * num_pixels;
	color_noise = new float[num_elems];
	ref_image = new unsigned char[num_elems];
	layers = new Spline_SimpleLayer*[max_layers];

	stroke_ref_img_ptr = NULL;

	for(int i = 0; i < num_elems; i++) {
		color_noise[i] = rand() * noise_gen_factor - 1.f;
	}


	for(int i = 0; i < max_layers; i++) {
		layers[i] = new Spline_SimpleLayer(width, height);
	}


	g_stroke_spacing = 0;
	g_stroke_lifetime = 10;
	g_stroke_fadetime = 10;
	current_layer = 0;
	bgfg_threshold = 28;
	render_start = 0;
	render_end = 0;
	QueryPerformanceFrequency((LARGE_INTEGER*)&timer_freq);

	initial_image_regeneration = new bool[max_layers];
	for(int i = 0; i < max_layers; i++) initial_image_regeneration[i] = false;


	_renderImage = NULL;


	offscreen = new RenderTexture(width, height);
	quad = new QuadVBO();
	tmp_target_RT = new RenderTexture(width, height);


#if USE_GLSL_SPLINE == 1
	// preload shader programs.
	GLSLManager::SetShader("Modulate.vert", "", "Modulate.frag");
	GLSLManager::SetShader("Height.vert", "", "Height0.frag");
	GLSLManager::SetShader("Height.vert", "", "Height1.frag");
	GLSLManager::SetShader();
#endif


	// set up timing environment
	d_spline_num_steps = 2 + max_layers * 4;
	d_spline_num_grows = 2 * max_layers;
	d_spline_num_refimgcomps = 3 * max_layers;
	d_spline_num_cvblur_cvcopy = 3 * max_layers;

#if D_USE_DEPTH_SPLINE == 0
	d_spline_num_listcomps = 4 * max_layers;
#else
	d_spline_num_listcomps = 3 * max_layers;
#endif

	d_spline_num_clears = 3;

	D_INIT_TIMING(d_spline_step, d_spline_num_steps);
	D_INIT_TIMING(d_spline_grows, d_spline_num_grows);
	D_INIT_TIMING(d_spline_refimgcomp, d_spline_num_refimgcomps);
	D_INIT_TIMING(d_spline_cvblur_cvcopy, d_spline_num_cvblur_cvcopy);
	D_INIT_TIMING(d_spline_listcomp, d_spline_num_listcomps);
	D_INIT_TIMING(d_spline_clear, d_spline_num_clears)
}


void SplineRenderer::finalizeBuffer() {
	delete[] color_noise;
	delete[] ref_image;

	for(int i = 0; i < max_layers; i++) {
		delete layers[i];
	}

	delete[] layers;
	delete[] initial_image_regeneration;

	delete offscreen;
	delete quad;
	delete tmp_target_RT;


	D_FIN_TIMING(d_spline_step)
	D_FIN_TIMING(d_spline_grows)
	D_FIN_TIMING(d_spline_refimgcomp)
	D_FIN_TIMING(d_spline_cvblur_cvcopy)
	D_FIN_TIMING(d_spline_listcomp)
	D_FIN_TIMING(d_spline_clear)
}


void SplineRenderer::setParams(IPaintingParam *param) {
	this->param = *static_cast<SplineParam*>(param);
}


bool SplineRenderer::render(unsigned char *result) {
	RenderTexture *prev_rt;

	if(result == NULL) return false;

	painted_image = result;
	render_start = 0;
	render_end = 0;

	for(int i = 0; i < max_layers; i++) initial_image_regeneration[i] = false;


	prev_rt = GLSLManager::GetCurrentRenderTarget();

	// set for this rendering
	GLSLManager::SetRenderTarget(offscreen);


	param.embossing_type = (int)::clampVal(param.embossing_type, 0, 1);
	hm_ptr = heightmap_handle[param.embossing_type];


	if(render_mode == 0) {
		_renderImage = &SplineRenderer::createLayeredImage;
	}
	else {
		_renderImage = &SplineRenderer::createHeightImage;
	}


	// select which image to use to look up stroke color
	stroke_ref_img_ptr = param.stroke_ref_img ? param.stroke_ref_img : ref_image;


	// render
	QueryPerformanceCounter((LARGE_INTEGER*)&render_start);

	(this->*_renderImage)();

	QueryPerformanceCounter((LARGE_INTEGER*)&render_end);


	// disconnect
	painted_image = NULL;
	_renderImage = NULL;
	hm_ptr = NULL;


	// restore
	if(prev_rt != NULL) {
		GLSLManager::SetRenderTarget(prev_rt);
	}
	else {
		GLSLManager::SetRenderTarget();
	}

	return true;
}


bool SplineRenderer::resize(int width, int height) {
	if(0 <= width || 0 <= height) return false;

	finalizeBuffer();

	this->width = width;
	this->height = height;

	initializeBuffer();

	return true;
}


bool SplineRenderer::setNumLayers(int num_layers) {
	return false;
}


void SplineRenderer::pushSeeds(int layer, const float *opt_flow, int width, int height) {
	int x, y, idx;
	int h_1 = this->height - 1;
	seed_list *ls;
	seed_list::iterator it, itr_end;
	SimpleStroke *stroke;

	if(opt_flow == NULL || width <= 0 || height <= 0) return;


	if(0 <= layer && layer < max_layers) {
		ls = &(layers[layer]->strokes);
		it = ls->begin();
		itr_end = ls->end();

		for(; it != itr_end; it++) {
			stroke = reinterpret_cast<SimpleStroke*>(*it);
			x = (int)stroke->x;
			y = (int)(h_1 - stroke->y);

			if(x < 0 || x >= this->width || y < 0 || y >= this->height) {	// should be within rendering window
				stroke->valid = false;
			}
			else if(x < width && y < height) {								// shift if it is within defined optical flow
				idx = 2 * (x + width * y);

				stroke->x = (float)(int)opt_flow[idx];
				stroke->y = (float)(int)(h_1 - opt_flow[idx + 1]);

				if(stroke->x < 0 || stroke->x >= this->width || stroke->y < 0 || stroke->y >= this->height) {
					stroke->valid = false;
				}
			}
		}
	}
}


size_t SplineRenderer::getNumStrokes(int layer) {
	if(0 <= layer && layer < max_layers) {
		return this->layers[layer]->strokes.size();
	}
	else {
		return 0;
	}
}


double SplineRenderer::getRenderingTime() {
	return (double)(render_end - render_start) / (double)timer_freq;
}


void SplineRenderer::clearSeeds(int layer_idx) {
	if(0 <= layer_idx && layer_idx < max_layers) {
		s_stroke_factory.dump(layers[layer_idx]->strokes);
	}
}



void SplineRenderer::makeStrokeTemplate() {
	// I use stroke images that are copied from Hertzmann's paper (Painterly Rendering with Curved Brush Strokes of Multiple Sizes)
	// You might need to replace stroke images or cite where you get them from, if you publish a peper...
	alphamap_handle = StrokeMapManager::makeStrokeShapeTexture("spline_stroke_alpha.bmp");

	heightmap_handle[0][0] = StrokeMapManager::makeStrokeHeightmapTexture("spline_stroke_height0_0.bmp");
	heightmap_handle[0][1] = StrokeMapManager::makeStrokeHeightmapTexture("spline_stroke_height0_1.bmp");
	heightmap_handle[0][2] = StrokeMapManager::makeStrokeHeightmapTexture("spline_stroke_height0_2.bmp");
	heightmap_handle[1][0] = StrokeMapManager::makeStrokeHeightmapTexture("spline_stroke_height1_0.bmp");
	heightmap_handle[1][1] = StrokeMapManager::makeStrokeHeightmapTexture("spline_stroke_height1_1.bmp");
	heightmap_handle[1][2] = StrokeMapManager::makeStrokeHeightmapTexture("spline_stroke_height1_2.bmp");

	hm_ptr = NULL;
}


void SplineRenderer::cleanSeedTemplate() {
	glDeleteTextures(1, &alphamap_handle);
	glDeleteTextures(1, &heightmap_handle[0][0]);
	glDeleteTextures(1, &heightmap_handle[0][1]);
	glDeleteTextures(1, &heightmap_handle[0][2]);
	glDeleteTextures(1, &heightmap_handle[1][0]);
	glDeleteTextures(1, &heightmap_handle[1][1]);
	glDeleteTextures(1, &heightmap_handle[1][2]);
}


void SplineRenderer::setUpMultiTexturePainting() {
#if USE_GLSL_SPLINE == 0
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_2D);
	glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[0]));
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
#endif
}


void SplineRenderer::cleanUpMultiTexturePainting() {
#if USE_GLSL_SPLINE == 0
	glActiveTexture(GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, prev_tex_env_mode[0]);
	glDisable(GL_TEXTURE_2D);
#endif
}



void SplineRenderer::setUpMultiTexturePostProcess() {
#if USE_GLSL_SPLINE == 0
	switch(param.embossing_type) {
		case 0:
			// stage 0: scale z-value of height map (scaling is applied in glSpline objects)
			glActiveTexture(GL_TEXTURE0);
			glEnable(GL_TEXTURE_2D);
			glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[0]));
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_RGB, GL_TEXTURE);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_RGB, GL_CONSTANT);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);

			// stage 1: add with offset (provided as a fragment color (primary color), specified in each glSpline object)
			glActiveTexture(GL_TEXTURE1);
			glEnable(GL_TEXTURE_2D);
			glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[1]));
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_ADD);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_RGB, GL_PREVIOUS);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_RGB, GL_PRIMARY_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);

			// stage 2: alpha-blend with stroke shape texture
			glActiveTexture(GL_TEXTURE2);
			glEnable(GL_TEXTURE_2D);
			glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[2]));
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

			break;


		case 1:
			// stage 0: scale z-value of height map (scaling is applied as tex. env. color)
			glActiveTexture(GL_TEXTURE0);
			glEnable(GL_TEXTURE_2D);
			glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[0]));
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_RGB, GL_TEXTURE);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_RGB, GL_CONSTANT);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);


			// stage 1: subtract shift from stroke color (primary color), add opacity
			glActiveTexture(GL_TEXTURE1);
			glEnable(GL_TEXTURE_2D);
			glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[1]));
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_SUBTRACT);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_RGB, GL_PRIMARY_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_RGB, GL_PREVIOUS);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_REPLACE);
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_ALPHA, GL_PRIMARY_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, GL_SRC_COLOR);


			// stage 2: alpha-blend with stroke shape texture
			glActiveTexture(GL_TEXTURE2);
			glEnable(GL_TEXTURE_2D);
			glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &(prev_tex_env_mode[2]));
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

			break;


		default:
			break;
	}

	glActiveTexture(GL_TEXTURE0);

#endif
}


void SplineRenderer::cleanUpMultiTexturePostProcess() {
#if USE_GLSL_SPLINE == 0
	glActiveTexture(GL_TEXTURE2);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, prev_tex_env_mode[2]);
	glDisable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE1);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, prev_tex_env_mode[1]);
	glDisable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, prev_tex_env_mode[0]);

	glDisable(GL_TEXTURE_2D);
#endif
}


void SplineRenderer::makeRefImage() {
	int idx1, idx2;
	int preset_idx = 0;
	int kernel_size = 1;
	int num_elems = 3 * num_pixels;
	
	IplImage *ipl_img = NULL;
	IplImage *ipl_ref = NULL;
	IplImage *ipl_mask = NULL;
	char *ipl_img_origin;
	
	PaintingStyleNode *psn = param.style->getStyleNode();
	int num_styles = (int)param.style->getNumRegions();

	map<int, PaintingRegion*>::iterator it, it_end;
	PaintingRegion *pr =NULL;
	int region_id;



	D_SET_COUNTER_VAR(d_cnt_idx, 3 * current_layer)
	D_SET_COUNTER_VAR(d_cnt_idx2, 0)


	D_TICK(d_spline_refimgcomp, d_cnt_idx)

	ipl_img = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 3);
	ipl_ref = cvCloneImage(ipl_img);
	ipl_mask = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 1);
	ipl_img_origin = ipl_img->imageDataOrigin;

	it_end = param.style->getEnd();

	D_TOCK(d_spline_refimgcomp, d_cnt_idx++)


	D_TICK(d_spline_refimgcomp, d_cnt_idx)

	// create a reference image by smoothing input. this image will be used in later
	// let sigma be computed from kernel_size
	for(it = param.style->getBegin(); it != it_end; it++) {

		D_UPDATE_COUNTER_VAR(d_cnt_idx2, 3 * current_layer)		// time is measured on only one region...
		D_TICK(d_spline_cvblur_cvcopy, d_cnt_idx2)

		pr = it->second;
		region_id = pr->getID();

		for(int j = 0; j < num_elems; j++) {
			ipl_img_origin[j] = (char)param.source_img[j];
		}

		kernel_size = 1;

		for(int j = 0; j < height; j++) {
			idx1 = width * j;
			idx2 = width * (height - 1 - j);

			for(int k = 0; k < width; k++) {
				if(region_id == psn[idx2].region_id) {
					ipl_mask->imageDataOrigin[idx1++] = 1;
					kernel_size = (int)(psn[idx2++].preset.stroke_size);
				}
				else {
					ipl_mask->imageDataOrigin[idx1++] = 0;
					idx2++;
				}
			}
		}

		// make it odd-size
		if(((kernel_size >> 1) << 1) == kernel_size) kernel_size++;

		D_TOCK(d_spline_cvblur_cvcopy, d_cnt_idx2++)

		D_TICK(d_spline_cvblur_cvcopy, d_cnt_idx2)
		cvSmooth(ipl_img, ipl_img, CV_GAUSSIAN, kernel_size, 0);		// this will be slow if kernel size is too large
		D_TOCK(d_spline_cvblur_cvcopy, d_cnt_idx2++)

		D_TICK(d_spline_cvblur_cvcopy, d_cnt_idx2)
		cvCopy(ipl_img, ipl_ref, ipl_mask);
		D_TOCK(d_spline_cvblur_cvcopy, d_cnt_idx2++)
	}

	D_TOCK(d_spline_refimgcomp, d_cnt_idx++)


	D_TICK(d_spline_refimgcomp, d_cnt_idx)

	// upside-down
	ipl_img_origin = ipl_ref->imageDataOrigin;

	for(int i = 0; i < height; i++) {
		idx1 = 3 * width * i;
		idx2 = 3 * width * (height - 1 - i);

		for(int j = 0; j < width; j++) {
			ref_image[idx1++] = (unsigned char)(ipl_img_origin[idx2++]);
			ref_image[idx1++] = (unsigned char)(ipl_img_origin[idx2++]);
			ref_image[idx1++] = (unsigned char)(ipl_img_origin[idx2++]);
		}
	}

	cvReleaseImage(&ipl_img);
	cvReleaseImage(&ipl_ref);
	cvReleaseImage(&ipl_mask);

	D_TOCK(d_spline_refimgcomp, d_cnt_idx++)
}


void SplineRenderer::regenerateStrokeList() {
	/* For now, assume that (width == height)... */
	int xi, yi;
	int w_1 = width - 1;
	int h_1 = height - 1;
	int psn_idx, idx_3;
	int push_id;
	int num_grids;
	int stroke_half_lifetme = g_stroke_lifetime + g_stroke_fadetime;
	float offset, center, delta;
	float half_rand = RAND_MAX * .5f;
	float decay_factor = 1.f;

	SimpleStroke *stroke = NULL;
	PaintingPreset *pp = NULL;
	vector< SimpleStroke* > seed_order;
	vector< SimpleStroke* >::iterator order_itr, order_end;

	PaintingStyleNode *psn = param.style->getStyleNode();
	FieldSampleImage *fsi = param.edge_pixel;
	FieldSample *fs = NULL;


#if D_USE_DEPTH_SPLINE == 0
	D_SET_COUNTER_VAR(d_cnt_idx, 4 * current_layer)
#else
	D_SET_COUNTER_VAR(d_cnt_idx, 3 * current_layer)
#endif

	D_TICK(d_spline_listcomp,  d_cnt_idx)

	clearSeeds(current_layer);

	for(int i = 0; i < current_layer; i++) {
		decay_factor *= .5f;
	}


	// skip if either the stroke size or grid size is too small
	if(g_stroke_spacing < 1.f) return;

	num_grids = (int)ceil((double)width / g_stroke_spacing);
	delta = g_stroke_spacing;
	offset = (num_grids * delta - width) * .5f;
	center = delta * .5f - offset;

	srand(0);

	D_TOCK(d_spline_listcomp,  d_cnt_idx++)

		
	D_TICK(d_spline_listcomp,  d_cnt_idx)

	if(current_layer == 0) {
		const float jitter_scale_factor = delta / (100.f * RAND_MAX);
		float jitter_scale;

		for(int i = 0; i < num_grids; i++) {
			for(int j = 0; j < num_grids; j++) {
				xi = (int)::clampVal(center + j * delta, 0, w_1);
				yi = (int)::clampVal(center + i * delta, 0, h_1);
				psn_idx = xi + width * yi;
				//idx_3 = 3 * psn_idx;

				pp = &(psn[psn_idx].preset);
				//fs = &(fsi->fld[xi + width * (h_1 - yi)]);


				jitter_scale = pp->stroke_pos_jitter * jitter_scale_factor;

				xi = (int)::clampVal(xi + (int)(jitter_scale * (rand() - half_rand)), 0, w_1);
				yi = (int)::clampVal(yi + (int)(jitter_scale * (rand() - half_rand)), 0, h_1);
				push_id = rand();


				psn_idx = xi + width * yi;
				idx_3 = 3 * psn_idx;

				pp = &(psn[psn_idx].preset);
				fs = &(fsi->fld[xi + width * (h_1 - yi)]);


				stroke = s_stroke_factory.make(0);
				stroke->x = (float)xi;
				stroke->y = (float)yi;
				stroke->size = pp->stroke_size * decay_factor;
				stroke->lifetime = stroke_half_lifetme;

				stroke->stroke_jitter_shift[0] = color_noise[idx_3];
				stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
				stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

				stroke->jitter_id = push_id % 3;
				stroke->depth = push_id;

				// alpha value will be computed at traceStrokes()
				//Spline_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
				Spline_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
				stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

				seed_order.push_back(stroke);
			}
		}
	}
	else {
		int cnt;
		int tmp_psn_idx;
		int grid_size = (int)delta;
		int num_kernel_elems;
		int *xs = NULL, *ys = NULL;
		int kx, ky, kx0, ky0;
		float *color_diff = new float[num_pixels];
		float area_error;

		// make odd-size
		if(((grid_size >> 1) << 1) == grid_size) grid_size++;

		num_kernel_elems = grid_size * grid_size;
		xs = new int[num_kernel_elems];
		ys = new int[num_kernel_elems];

		// compute color difference between painted image and reference image
		computeColorDifference(painted_image, ref_image, width, height, color_diff);

		//if(current_layer == 1)saveFloatImg(width, height, 1, color_diff, _T("floatImgSpline.ppm"));

		for(int i = 0; i < num_grids; i++) {
			for(int j = 0; j < num_grids; j++) {
				xi = (int)::clampVal(center + j * delta, 0, w_1);
				yi = (int)::clampVal(center + i * delta, 0, h_1);
				psn_idx = xi + width * yi;

				push_id = rand();

				if(current_layer >= psn[psn_idx].preset.num_layers) continue;

				area_error = 0;
				cnt = 0;

				findRegion(width, height, xi, yi, grid_size, xs, ys, false);

				for(int m = 0; m < num_kernel_elems; m++) {
					kx = xs[m];
					ky = ys[m];

					if(0 <= kx && kx < width && 0 <= ky && ky < height) {
						area_error += color_diff[kx + width * ky];
						cnt++;
					}
				}

				if(cnt > 0) {
					area_error /= cnt;

					if(psn[psn_idx].preset.area_error < area_error) {
						psn_idx = -1;

						// find a pixel with the largest error
						for(int m = 0; m < num_kernel_elems; m++) {
							kx = xs[m];
							ky = ys[m];

							if(0 <= kx && kx < width && 0 <= ky && ky < height) {
								tmp_psn_idx = kx + width * ky;

								if(psn_idx < 0 || color_diff[tmp_psn_idx] > color_diff[psn_idx]) {
									psn_idx = tmp_psn_idx;
									kx0 = kx;
									ky0 = ky;
									//break;
								}
							}
						}

						if(psn_idx >= 0) {
							idx_3 = 3 * psn_idx;
							pp = &(psn[psn_idx].preset);
							fs = &(fsi->fld[kx0 + width * (h_1 - ky0)]);

							stroke = s_stroke_factory.make(0);
							stroke->x = (float)kx0;
							stroke->y = (float)ky0;
							stroke->size = pp->stroke_size * decay_factor;
							stroke->lifetime = stroke_half_lifetme;

							stroke->stroke_jitter_shift[0] = color_noise[idx_3];
							stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
							stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

							stroke->jitter_id = push_id % 3;
							stroke->depth = push_id;

							// alpha value will be computed at traceStrokes()
							//Spline_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							Spline_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

							seed_order.push_back(stroke);
						}
					}
				}
			}
		}

		delete[] xs;
		delete[] ys;
		delete[] color_diff;
	}

	D_TOCK(d_spline_listcomp,  d_cnt_idx++)



#if D_USE_DEPTH_SPLINE == 0

	D_TICK(d_spline_listcomp,  d_cnt_idx)

	switch(param.seed_order) {
		case 0:
		default:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeDepth);
			break;

		case 1:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeRGB);
			break;

		case 2:
			// extra step to compute hsv color...
			order_itr = seed_order.begin();
			order_end = seed_order.end();

			for(; order_itr != order_end; order_itr++) {
				RgbHsv((*order_itr)->color, (*order_itr)->hsv);
			}

			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeHSV);
			break;
	}

	D_TOCK(d_spline_listcomp,  d_cnt_idx++)


	D_TICK(d_spline_listcomp,  d_cnt_idx)

	// create strokes
	for(size_t i = 0; i < seed_order.size(); i++) {
		layers[current_layer]->strokes.push_back(seed_order[i]);
	}

	D_TOCK(d_spline_listcomp,  d_cnt_idx++)

#else

	D_TICK(d_spline_listcomp,  d_cnt_idx)

	switch(param.seed_order) {
		case 0:
		default:
			for(size_t i = 0; i < seed_order.size(); i++) {
				seed_order[i]->z = seed_order[i]->depth / (float)RAND_MAX;
				layers[current_layer]->strokes.push_back(seed_order[i]);
			}

			break;

		case 1:
			for(size_t i = 0; i < seed_order.size(); i++) {
				seed_order[i]->z = calcDepth(seed_order[i]->color);
				layers[current_layer]->strokes.push_back(seed_order[i]);
			}

			break;
	}

	D_TOCK(d_spline_listcomp,  d_cnt_idx++)

#endif

}




void SplineRenderer::updateStrokeList() {
	/* For now, assume that (width == height)... */
	int xi, yi;
	int w_1 = width - 1;
	int h_1 = height - 1;
	int psn_idx, idx_3;
	int push_id;
	int stroke_half_lifetime = g_stroke_fadetime + g_stroke_lifetime;
	int stroke_total_lifetime = 2 * g_stroke_fadetime + g_stroke_lifetime;
	int num_grids;
	int removal_cnt[2] = {0, 0};
	float offset, delta;
	SimpleStroke *stroke;
	PaintingPreset *pp = NULL;
	vector< SimpleStroke* > seed_order;
	vector< SimpleStroke* >::iterator order_itr, order_end;


	PaintingStyleNode *psn = param.style->getStyleNode();


	seed_list *ls;
	seed_list::iterator it, itr_end;

	unsigned char *stroke_map = NULL;

	int num_subgrids;
	int hs;
	int cover;
	int tx, ty;
	bool flag;


	FieldSampleImage *fsi = param.edge_pixel;
	FieldSample *fs = NULL;


	float min_stroke_size = param.style->findMinStrokeSize(param.time);
	hs = ((int)min_stroke_size >> 1) + 1;
	cover = 2 * hs;


	//----------------------------------------------------------------------------------------------------------//
	//		Compute stroke map
	//----------------------------------------------------------------------------------------------------------//


	// remove all invalid strokes that are pushed outside of canvas
	ls = &(layers[current_layer]->strokes);
	it = ls->begin();
	itr_end = ls->end();

	if(current_layer == 0) {
		for(; it != itr_end; ) {
			if((*it)->valid == false) {
				s_stroke_factory.dump(*it);
				it = ls->erase(it);
				removal_cnt[0]++;
			}
			else {
				it++;
			}
		}
	}
	else {
		for(; it != itr_end; ) {
			(*it)->lifetime--;

			if((*it)->lifetime <= 0 || (*it)->valid == false) {
				s_stroke_factory.dump(*it);
				it = ls->erase(it);
				removal_cnt[0]++;
			}
			else {
				it++;
			}
		}
	}


	// return here
	if(min_stroke_size <= 0) return;


	// merge crowded seeds
	//...how to deal with the transparency?
	int bin_size, sqrt_bin_size;
	float d, ood;
	int containment_threshold = 1;
	vector<SimpleStroke*> *s_bin;
	vector<SimpleStroke*>::iterator itv;

	if(current_layer == 0) {
		d = (float)floor(min_stroke_size / (2 * 1.42));
		
		if(d <= 0) d = 1;
	}
	else{
		d = g_stroke_spacing;
	}

	ood = 1 / d;

	sqrt_bin_size = (int)ceil((float)width * ood);
	bin_size = sqrt_bin_size * sqrt_bin_size;
	s_bin = new vector<SimpleStroke*>[bin_size];


	containment_threshold = (int)ceil(d / ((min_stroke_size < 1) ? 1.f : min_stroke_size));
	containment_threshold *= containment_threshold;

	// We want to hold more stroke seeds that have transparency than those with 0 opacity..
	// Here comes a factor to scale the containment threshold...
	//
	// Factor is defined by 255 / (stroke_alpha)
	//
	// For simplicity, assume stroke opacity doesn't go below 128... 
	//
	containment_threshold *= 2;	
	itr_end = ls->end();

	for(it = ls->begin(); it != itr_end; it++) {
		xi = (int)::clampVal((*it)->x * ood, 0, sqrt_bin_size - 1);
		yi = (int)::clampVal((*it)->y * ood, 0, sqrt_bin_size - 1);

		s_bin[xi + sqrt_bin_size * yi].push_back(*it);
	}


	for(int i = 0; i < bin_size; i++) {
		if(s_bin[i].size() > (size_t)containment_threshold) {

#if D_USE_DEPTH_SPLINE == 1
			std::sort(s_bin[i].begin(), s_bin[i].end(), ::compareStrokeDepth);
#endif 
			for(size_t i = containment_threshold; i < s_bin[i].size(); i++) {
				s_bin[i].at(i)->valid = false;
			}
		}
	}


	itr_end = ls->end();

	for(it = ls->begin(); it != itr_end; ) {
		if((*it)->valid == false) {
			s_stroke_factory.dump(*it);
			it = ls->erase(it);
			removal_cnt[1]++;
		}
		else {
			it++;
		}
	}


	// recompute colors and angles (orientation)
	itr_end = ls->end();

	for(it = ls->begin(); it != itr_end; it++) {
		psn_idx = ((int)(*it)->x) + width * ((int)(*it)->y);
		idx_3 = 3 * psn_idx;
		pp = &(psn[psn_idx].preset);
		//Spline_sampleColorJitter((*it)->color, &(ref_image[idx_3]), pp->stroke_color_shift, (*it)->stroke_jitter_shift, pp->jitter_amplitude);
		Spline_sampleColorJitter((*it)->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, (*it)->stroke_jitter_shift, pp->jitter_amplitude);

		fs = &(fsi->fld[((int)(*it)->x) + width * (h_1 - ((int)(*it)->y))]);

		if(cos((*it)->angle) * fs->vec_fld.entry[0] + sin((*it)->angle) * fs->vec_fld.entry[1] < 0) {
			(*it)->angle = (float)atan2(-1 * fs->vec_fld.entry[1], -1 * fs->vec_fld.entry[0]);
		}
		else {
			(*it)->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);
		}
	}


	delete[] s_bin;



	if(g_stroke_spacing < 1.f) return;

	stroke_map = new unsigned char[3 * num_pixels];



	// make sure if client state for vertex array is enabled
	/*
	GLboolean enabled;
	glGetBooleanv(GL_VERTEX_ARRAY, &enabled);
	assert(enabled == GL_TRUE);
	glGetBooleanv(GL_TEXTURE_COORD_ARRAY, &enabled);
	assert(enabled == GL_TRUE);
	*/

	spline_list tmp_splines;
	float scale = 1.f;
	const float decay_factor = .5f;

	for(int i = 0; i < current_layer; i++) scale *= decay_factor;


	// create a temporary buffer to render a storoke map
	GLSLManager::SetRenderTarget(tmp_target_RT);
	glClearColor(1.f, 1.f, 1.f, 1.f);
	//glClearColor(1.f, 1.f, 1.f, 0.f);
	glClear(GL_COLOR_BUFFER_BIT);
	traceStrokes(*ls, tmp_splines, scale, 1);
	renderTexturedStrokes(tmp_splines);
	glReadPixels(0, 0, width, height, GL_RED, GL_UNSIGNED_BYTE, stroke_map);
	//glReadPixels(0, 0, width, height, GL_ALPHA, GL_UNSIGNED_BYTE, stroke_map);
	GLSLManager::SetRenderTarget(offscreen);

	/**
	if(current_layer == 1) {
		int idx = 0;
		unsigned char *bg = new unsigned char[3 * num_pixels];
		for(int i = 0; i < num_pixels; i++, idx += 3) {
			bg[idx] = stroke_map[i];
			bg[idx + 1] = stroke_map[i];
			bg[idx + 2] = stroke_map[i];
		}
		My_SSS_SavePPM(width, height, bg, _T("spline_strokemap_spline.ppm"));
		delete[] bg;
	}
	/**/

	spline_factory.dump(tmp_splines);		// reuse



	//----------------------------------------------------------------------------------------------------------//
	//		Find where to put new seeds
	//----------------------------------------------------------------------------------------------------------//
	num_grids = (int)ceil((double)width / g_stroke_spacing);
	offset = (num_grids *  g_stroke_spacing - width) * .5f;

	srand(0);

	if(current_layer == 0) {
		float corner;
		delta = (float)floor(min_stroke_size / (2 * 1.42));

		if(delta <= 0) delta = 1;

		num_subgrids = (int)ceil(width / (double)delta);

		// it's not exactly a center at all... rather, it is a corner position.
		corner = (width - num_subgrids * delta) * .5f;

		for(int i = 0; i < num_subgrids; i++) {
			for(int j = 0; j < num_subgrids; j++) {
				xi = (int)(corner + j * delta);
				yi = (int)(corner + i * delta);

				push_id = rand();
				flag = false;

				for(int m = 0; m < delta && flag == false; m++) {
					ty = yi + m;

					if(0 <= ty && ty < height) {
						for(int n = 0; n < delta && flag == false; n++) {
							tx = xi + n;

							if(0 <= tx && tx < width) {
								if(stroke_map[tx + width * ty] > bgfg_threshold) {
									flag = true;

									// seed a stroke at first empty pixel
									xi = tx;
									yi = ty;
									psn_idx = xi + width * yi;
									idx_3 = 3 * psn_idx;
									pp = &(psn[psn_idx].preset);
									fs = &(fsi->fld[xi + width * (h_1 - yi)]);

									stroke = s_stroke_factory.make(0);
									stroke->x = (float)xi;
									stroke->y = (float)yi;
									stroke->size = pp->stroke_size * decay_factor;
									stroke->lifetime = stroke_half_lifetime;

									stroke->stroke_jitter_shift[0] = color_noise[idx_3];
									stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
									stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

									stroke->jitter_id = push_id % 3;
									stroke->depth = push_id;

									//Spline_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
									Spline_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
									stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

									seed_order.push_back(stroke);
								}
							}
						}
					}
				}
			}
		}
	}
	else {
		int cnt;
		int tmp_psn_idx;
		int grid_size;
		int num_kernel_elems;
		int *xs = NULL, *ys = NULL;
		int kx, ky, kx0, ky0;
		float *color_diff = new float[num_pixels];
		float area_error;
		float center;


		delta = g_stroke_spacing;
		num_subgrids = num_grids;		// won't be used...
		center = delta * .5f - offset;
		grid_size = (int)delta;


		// make odd-size
		if(((grid_size >> 1) << 1) == grid_size) grid_size++;

		num_kernel_elems = grid_size * grid_size;
		xs = new int[num_kernel_elems];
		ys = new int[num_kernel_elems];

		// compute color difference between painted image and reference image
		computeColorDifference(painted_image, ref_image, width, height, color_diff);

		for(int i = 0; i < num_grids; i++) {
			for(int j = 0; j < num_grids; j++) {
				xi = (int)::clampVal(center + j * delta, 0, w_1);
				yi = (int)::clampVal(center + i * delta, 0, h_1);
				psn_idx = xi + width * yi;

				push_id = rand();

				area_error = 0;
				cnt = 0;

				findRegion(width, height, xi, yi, grid_size, xs, ys, false);

				for(int m = 0; m < num_kernel_elems; m++) {
					kx = xs[m];
					ky = ys[m];

					if(0 <= kx && kx < width && 0 <= ky && ky < height) {
						area_error += color_diff[kx + width * ky];
						cnt++;
					}
				}

				if(cnt > 0) {
					area_error /= cnt;

					if(psn[psn_idx].preset.area_error < area_error) {
						psn_idx = -1;

						// find a pixel with the largest error
						for(int m = 0; m < num_kernel_elems; m++) {
							kx = xs[m];
							ky = ys[m];

							if(0 <= kx && kx < width && 0 <= ky && ky < height) {
								tmp_psn_idx = kx + width * ky;

								if(psn_idx < 0 || color_diff[tmp_psn_idx] > color_diff[psn_idx]) {
									psn_idx = tmp_psn_idx;
									kx0 = kx;
									ky0 = ky;
								}
							}
						}

						if(psn_idx >= 0 && stroke_map[psn_idx] > bgfg_threshold && current_layer < psn[psn_idx].preset.num_layers) {
							idx_3 = 3 * psn_idx;
							pp = &(psn[psn_idx].preset);
							fs = &(fsi->fld[kx0 + width * (h_1 - ky0)]);

							stroke = s_stroke_factory.make(0);
							stroke->x = (float)kx0;
							stroke->y = (float)ky0;
							stroke->size = pp->stroke_size * decay_factor;
							//stroke->lifetime = stroke_half_lifetime;
							stroke->lifetime = stroke_total_lifetime;


							stroke->stroke_jitter_shift[0] = color_noise[idx_3];
							stroke->stroke_jitter_shift[1] = color_noise[idx_3 + 1];
							stroke->stroke_jitter_shift[2] = color_noise[idx_3 + 2];

							stroke->jitter_id = push_id % 3;
							stroke->depth = push_id;

							//Spline_sampleColorJitter(stroke->color, &(ref_image[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							Spline_sampleColorJitter(stroke->color, &(stroke_ref_img_ptr[idx_3]), pp->stroke_color_shift, stroke->stroke_jitter_shift, pp->jitter_amplitude);
							stroke->angle = (float)atan2(fs->vec_fld.entry[1], fs->vec_fld.entry[0]);

							seed_order.push_back(stroke);
						}
					}
				}
			}
		}

		delete[] xs;
		delete[] ys;
		delete[] color_diff;
	}


	ls = &(layers[current_layer]->strokes);
	itr_end = ls->end();


	// copy over for sort a whole set of seeds...
	if(param.partial_ordering == false) {
		for(it = ls->begin(); it != itr_end; it++) {
			seed_order.push_back(*it);
		}

		ls->clear();
	}


	switch(param.seed_order) {
		case 0:
		default:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeDepth);
			break;

		case 1:
			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeRGB);
			break;

		case 2:
			// extra step to compute hsv color...
			order_itr = seed_order.begin();
			order_end = seed_order.end();

			for(; order_itr != order_end; order_itr++) {
				RgbHsv((*order_itr)->color, (*order_itr)->hsv);
			}

			std::sort(seed_order.begin(), seed_order.end(), StrokeComparator::compareStrokeHSV);
			break;
	}



	for(size_t i = 0; i < seed_order.size(); i++) {
		ls->push_back(seed_order[i]);
	}

	delete[] stroke_map;
}


void SplineRenderer::renderTexturedStrokes(spline_list &splines) {
	spline_list::reverse_iterator itr = splines.rbegin();
	spline_list::reverse_iterator end = splines.rend();


#if USE_GLSL_SPLINE == 1
	GLSLManager::SetShader("Modulate.vert", "", "Modulate.frag");
#else
	setUpMultiTexturePainting();
#endif

	for(; itr != end; itr++) {
		(*itr)->draw();
	}


#if USE_GLSL_SPLINE == 1
	GLSLManager::SetShader();
#else
	cleanUpMultiTexturePainting();
#endif

}


void SplineRenderer::createLayeredImage() {
	int idx = 0;
	int num_layers;
	int num_elems = 4 * num_pixels;

	const float decay_factor = .5;
	float scale = 1.f;

	PaintingStyleNode *psn = param.style->getStyleNode();



	D_SET_COUNTER_VAR(d_stepcount, 0)

	// initialization
	D_TICK(d_spline_step, d_stepcount)

	num_layers = param.style->getMaxLayers();

	// remove seeds on layers that won't be used for painting
	for(int i = num_layers; i < max_layers; i++) {
		clearSeeds(i);
	}

	// remove all tesselated strokes (they'll be re-generated)
	for(int i = 0; i < max_layers; i++) {
		spline_factory.dump(layers[i]->splines);
	}

	// global stroke parameter setting...
	if(param.style->getNumRegions() == 1) {
		g_stroke_spacing = psn[0].preset.stroke_gridsize;
	}
	else {
		g_stroke_spacing = param.style->getStrokeSpacingForMultiStyle();
	}



	glClearColor(1.f, 1.f, 1.f, 0.f);		// set the white canvas


#if D_USE_DEPTH_SPLINE == 0
	glClear(GL_COLOR_BUFFER_BIT);

#else
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.8);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

#endif

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, painted_image);

	D_TOCK(d_spline_step, d_stepcount++)


	for(int i = 0; i < num_layers; i++) {
		current_layer = i;

		// I assume that reference image is already available for height-map rendering... (i.e. the color-painting is already generated)
		D_TICK(d_spline_step, d_stepcount)

		if(render_mode == 0) {
			makeRefImage();
		}

		D_TOCK(d_spline_step, d_stepcount++)



		// create stroke list
		D_TICK(d_spline_step, d_stepcount)

		if(seed_gen_mode == 0) {
			regenerateStrokeList();
		}
		else {
			updateStrokeList();
		}

		D_TOCK(d_spline_step, d_stepcount++)



		// grow splines
		D_TICK(d_spline_step, d_stepcount)

		traceStrokes(layers[current_layer]->strokes, layers[current_layer]->splines, scale);

		D_TOCK(d_spline_step, d_stepcount++)



		// render strokes (Hertzmann avoids overhead of storing/sorting strokes so he uses the Z-buffer... 
		// but we don't do it here)
		D_TICK(d_spline_step, d_stepcount)

		renderTexturedStrokes(layers[current_layer]->splines);

		scale *= decay_factor;

		glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, painted_image);

		D_TOCK(d_spline_step, d_stepcount++)
	}


	// clear up rendering
	D_UPDATE_COUNTER_VAR(d_stepcount, 1 + 4 * max_layers)
	D_TICK(d_spline_step, d_stepcount)
	D_TICK(d_spline_clear, 0)

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);

#if D_USE_DEPTH_SPLINE == 1
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_ALPHA_TEST);
#endif

	D_TOCK(d_spline_clear, 0)
	D_TICK(d_spline_clear, 1)

	// throw extra strokes
	spline_factory.hold();

	D_TOCK(d_spline_clear, 1)
	D_TICK(d_spline_clear, 2)

	// throw extra seeds
	s_stroke_factory.hold();

	D_TOCK(d_spline_clear, 2)
	D_TOCK(d_spline_step, d_stepcount++)



#if D_SAVESTAT_SPLINE == 1

	string path(VisView::getFullPath());
	path.append("\\Spline_stat.txt");
	FILE *d_fp = fopen(path.c_str(), "a");

	int d_step_idx = 0;
	int d_grow_idx = 0;
	int d_refimgcmp_idx = 0;
	int d_listcomp_idx = 0;
	int d_cvblur_copy_idx = 0;
	int d_clear_idx = 0;

	fprintf(d_fp, "Image rendering:                                         \tTime(sec)\n");
	fprintf(d_fp, "Initialize (clear seeds in unused layers, clear strokes):\t%.3f\n\n", D_TIME(d_spline_step, d_step_idx, timer_freq)); d_step_idx++;

	for(int i = 0; i < num_layers; i++) {
		fprintf(d_fp, "Layer %d\n", i);
		fprintf(d_fp, "  Make ref. image:                                       \t\n", D_TIME(d_spline_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "    Initialize:                                          \t%.3f\n", D_TIME(d_spline_refimgcomp, d_refimgcmp_idx, timer_freq)); d_refimgcmp_idx++;
		fprintf(d_fp, "    Blur:                                                \t\n", D_TIME(d_spline_refimgcomp, d_refimgcmp_idx, timer_freq)); d_refimgcmp_idx++;
		fprintf(d_fp, "      Setup:                                             \t%.3f\n", D_TIME(d_spline_cvblur_cvcopy, d_cvblur_copy_idx, timer_freq)); d_cvblur_copy_idx++;
		fprintf(d_fp, "      cvBlur():                                          \t%.3f\n", D_TIME(d_spline_cvblur_cvcopy, d_cvblur_copy_idx, timer_freq)); d_cvblur_copy_idx++;
		fprintf(d_fp, "      cvCopy():                                          \t%.3f\n", D_TIME(d_spline_cvblur_cvcopy, d_cvblur_copy_idx, timer_freq)); d_cvblur_copy_idx++;
		fprintf(d_fp, "    Copy & finalize:                                     \t%.3f\n", D_TIME(d_spline_refimgcomp, d_refimgcmp_idx, timer_freq)); d_refimgcmp_idx++;

		fprintf(d_fp, "  Compute a seed list:                                   \t\n", D_TIME(d_spline_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "    Initialize:                                          \t%.3f\n", D_TIME(d_spline_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;
		fprintf(d_fp, "    Generate seeds:                                      \t%.3f\n", D_TIME(d_spline_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;
#if D_USE_DEPTH_SPLINE == 0
		fprintf(d_fp, "    Sort:                                                \t%.3f\n", D_TIME(d_spline_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;
#endif
		fprintf(d_fp, "    Assign depth:                                        \t%.3f\n", D_TIME(d_spline_listcomp, d_listcomp_idx, timer_freq)); d_listcomp_idx++;

		fprintf(d_fp, "  Grow/Tesselate:                                        \t%\n", D_TIME(d_spline_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "    Grow:                                                \t%.3f\n", D_TIME(d_spline_grows, d_grow_idx, timer_freq)); d_grow_idx++;
		fprintf(d_fp, "    Tesselate:                                           \t%.3f\n", D_TIME(d_spline_grows, d_grow_idx, timer_freq)); d_grow_idx++;

		fprintf(d_fp, "  Render:                                                \t%.3f\n", D_TIME(d_spline_step, d_step_idx, timer_freq)); d_step_idx++;
		fprintf(d_fp, "\n");
	}

	d_step_idx = 1 + max_layers * 4;
	fprintf(d_fp, "Clean up:                                                \t%.3f\n", D_TIME(d_spline_step, d_step_idx, timer_freq)); d_step_idx++;
	fprintf(d_fp, "  Disable GL states:                                     \t%.3f\n", D_TIME(d_spline_clear, d_clear_idx, timer_freq)); d_clear_idx++;
	fprintf(d_fp, "  Release memory for extra strokes:                      \t%.3f\n", D_TIME(d_spline_clear, d_clear_idx, timer_freq)); d_clear_idx++;
	fprintf(d_fp, "  Release memory for extra seeds:                        \t%.3f\n", D_TIME(d_spline_clear, d_clear_idx, timer_freq)); d_clear_idx++;
	fprintf(d_fp, "---------------------------------------------------------------\n");
	fprintf(d_fp, "Render Time:                                             \t%.3f\n", (double)(d_spline_step[1][d_step_idx - 1]-d_spline_step[0][0]) / (double)timer_freq);
	fprintf(d_fp, "Number of strokes: [");

	for(int i = 0; i < max_layers; i++) {
		fprintf(d_fp, "%d", getNumStrokes(i));

		if(i < max_layers - 1) fprintf(d_fp, ", ");
	}

	fprintf(d_fp, "]\n");
	fprintf(d_fp, "\n\n");
	fclose(d_fp);

#endif
}


void SplineRenderer::traceStrokes(seed_list &strokes, spline_list &splines, float scale, int color_mode, bool render_initial_seeds) {
	int min_itr, max_itr;
	int h_1 = height - 1;
	int d0, d1;
	int start_region_id;
	unsigned char color_ub[3];
	float color_f[4] = {0, 0, 0, 1.f};
	float frac = 1 / 255.f;
	float sf = 1.f / width;		// assume width == height
	float stroke_radius;
	//float scaled_stroke_falloff = g_stroke_falloff / 100.f;
	float scaled_stroke_falloff = 0;
	seed_list::iterator itr, end;
	GLSpline *sp = NULL;
	PaintingPreset *pp;

	FieldSample *b_fs = NULL;
	FieldSampleImage *fsi = param.edge_pixel;
	PaintingStyleNode *fsn = param.style->getStyleNode();

	int b_xi, b_yi, b_zi = 0;
	double b_xd, b_yd, b_zd;
	int cnt;
	double b_dx = 0, b_dy = 0;
	double b_l_dx = 0, b_l_dy = 0;

	int b_idx1, b_idx2;
	int b_idx_3;

	double inv_len = 0;
	double step_size = 0;

	int d_cnt_idx = 2 * current_layer;


	// stroke grow
	if(render_initial_seeds == false) { D_TICK(d_spline_grows, d_cnt_idx) }

	if(render_initial_seeds == false) {
		itr = strokes.begin();
		end = strokes.end();

		for(; itr != end; itr++) {
			b_xi = (int)(*itr)->x;
			b_yi = (int)(*itr)->y;
			b_l_dx = cos((*itr)->angle);
			b_l_dy = sin((*itr)->angle);
			b_idx1 = b_xi + width * b_yi;
			b_idx2 = b_xi + width * (h_1 - b_yi);
			b_idx_3 = 3 * b_idx1;
			b_fs = &(fsi->fld[b_idx2]);
			pp = &(fsn[b_idx1].preset);
			min_itr = pp->stroke_min_length;
			max_itr = pp->stroke_max_length;
			start_region_id = fsn[b_idx1].region_id;
	
			// convert xi & yi to xf & yf.
			gluUnProject(b_xi, b_yi, b_zi, param.modelview, param.projection, param.viewport, &b_xd, &b_yd, &b_zd);

			switch(color_mode) {
				case 0:
					color_ub[0] = (*itr)->color[0];
					color_ub[1] = (*itr)->color[1];
					color_ub[2] = (*itr)->color[2];

					break;

				case 1:
					color_ub[0] = 0;
					color_ub[1] = 0;
					color_ub[2] = 0;

					break;
			}



			color_f[0] = color_ub[0] * frac;
			color_f[1] = color_ub[1] * frac;
			color_f[2] = color_ub[2] * frac;


			color_f[3] = calcAlpha(pp->stroke_initial_opacity * frac, scaled_stroke_falloff, g_stroke_lifetime, g_stroke_fadetime, (*itr)->lifetime);

			// compute width of a stroke first
			stroke_radius = pp->stroke_size * scale;

			sp = spline_factory.make((float)b_xd, (float)b_yd, stroke_radius * sf, color_f, (*itr)->angle, true);
			sp->setAlphaTexture(alphamap_handle);
			sp->setDepth((*itr)->z);

			// assign height texture later...
			//sp->setHeightMapTexture(hm_ptr[(*itr)->jitter_id]);

			// then radius of the stroke...
			stroke_radius *= .5f;

			cnt = 1;

			while(cnt <= max_itr) {
				// compare underlying color
				d0 = computeSquaredColorDifference(&(ref_image[b_idx_3]), &(painted_image[b_idx_3]));
				d1 = computeSquaredColorDifference(&(ref_image[b_idx_3]), color_ub);

				if(cnt > min_itr && d0 < d1) {
					break;
				}

				// detect vanishing point (... is it better to use a weak gradient threshold??)
				if(b_fs->vec_fld.entry[0] == 0 && b_fs->vec_fld.entry[1] == 0) {
					break;
				}
	
				// check whether the control point is within the same region.
				if(start_region_id != fsn[b_idx1].region_id) {
					break;
				}

	
				b_dx = b_fs->vec_fld.entry[0];
				b_dy = b_fs->vec_fld.entry[1];
				inv_len = 1.f / sqrt(b_dx * b_dx + b_dy * b_dy);

				if(b_dx * b_l_dx + b_dy * b_l_dy < 0) {
					b_dx *= -1;
					b_dy *= -1;
				}

				// No direction filter...
				step_size = stroke_radius * inv_len;
				b_xi += (int)(step_size * b_dx);
				b_yi += (int)(step_size * b_dy);

				// add to a stroke list
				gluUnProject(b_xi, b_yi, b_zi, param.modelview, param.projection, param.viewport, &b_xd, &b_yd, &b_zd);
				sp->addBack((float)b_xd, (float)b_yd);

				b_l_dx = b_dx;
				b_l_dy = b_dy;

				// get out of loop if the new point is outside of image region
				if(b_xi < 0 || b_xi >= width || b_yi < 0 || b_yi >= height) {
					break;
				}
				else {
					b_idx1 = b_xi + width * b_yi;
					b_idx2 = b_xi + width * (h_1 - b_yi);
					b_idx_3 = 3 * b_idx1;
					b_fs = &(fsi->fld[b_idx2]);
				}

				cnt++;
			}

			splines.push_back(sp);
		}
	}
	else {
		itr = strokes.begin();
		end = strokes.end();

		for(; itr != end; itr++) {
			b_xi = (int)(*itr)->x;
			b_yi = (int)(*itr)->y;
			b_idx1 = b_xi + width * b_yi;
			pp = &(fsn[b_idx1].preset);

			// convert xi & yi to xf & yf.
			gluUnProject(b_xi, b_yi, b_zi, param.modelview, param.projection, param.viewport, &b_xd, &b_yd, &b_zd);

			switch(color_mode) {
				case 0:
					color_ub[0] = (*itr)->color[0];
					color_ub[1] = (*itr)->color[1];
					color_ub[2] = (*itr)->color[2];

					break;

				case 1:
					color_ub[0] = 0;
					color_ub[1] = 0;
					color_ub[2] = 0;

					break;
			}

			color_f[0] = color_ub[0] * frac;
			color_f[1] = color_ub[1] * frac;
			color_f[2] = color_ub[2] * frac;
			color_f[3] = calcAlpha(pp->stroke_initial_opacity * frac, scaled_stroke_falloff, g_stroke_lifetime, g_stroke_fadetime, (*itr)->lifetime);

			// compute width of a stroke first
			stroke_radius = pp->stroke_size * scale;

			sp = spline_factory.make((float)b_xd, (float)b_yd, stroke_radius * sf, color_f, (*itr)->angle, true);
			sp->setAlphaTexture(alphamap_handle);
			sp->setDepth((*itr)->z);

			splines.push_back(sp);
		}
	}

	if(render_initial_seeds == false) { D_TOCK(d_spline_grows, d_cnt_idx++) }


	// tessellation
	if(render_initial_seeds == false) { D_TICK(d_spline_grows, d_cnt_idx) }

	spline_list::iterator sl_itr, sl_end;
	sl_end = splines.end();

	// stroke primitive tesselation
	int stroke_size = (int)splines.size();
	spline_list::iterator curr_itr = splines.begin();

	for(sl_itr = splines.begin(); sl_itr != sl_end; sl_itr++) {
		(*sl_itr)->tesselate(true);
	}

	if(render_initial_seeds == false) { D_TOCK(d_spline_grows, d_cnt_idx++) }
}


void SplineRenderer::createHeightImage() {
	int cnt = 0;
	int num_strokes = 0;
	size_t n_splines, n_strokes;
	bool err_flag = false;
	float offset, dh;
	float min_offset = .2f;
	float max_offset = .4f;
	const float frac = 1.f / 255.f;
	spline_list::reverse_iterator spline_itr, spline_end;
	seed_list::reverse_iterator stroke_itr;
	GLuint program_id, jitter_scale_loc, jitter_freq_loc;
	PaintingStyleNode *fsn = param.style->getStyleNode();
	PaintingPreset *pp;
	spline_list *splines = new spline_list[max_layers];


	int d_cnt_idx = 0;
	__int64 **mt;
	D_INIT_TIMING(mt, 2 + max_layers)


	// initialization
	D_TICK(mt, d_cnt_idx)

	for(int i = 0; i < max_layers; i++) {
		splines[i] = layers[i]->splines;
	}


	// I use a spline list for rendering and a stroke list to fetch a seed position in "pixels".
	// I also assume that spline and seed lists have the same number of elements...
	for(int i = 0; i < max_layers; i++) {
		n_splines = layers[i]->splines.size();
		n_strokes = layers[i]->strokes.size();

		if(n_splines == n_strokes) {
			num_strokes += (int)n_strokes;
		}
		else {
			err_flag = true;
			break;
		}
	}


	if(err_flag == true) {
		FILE *fp = fopen("SplineRenderer_error.txt", "w");
		fprintf(fp, "Incosistent number of seeds and splines...\r\n");
		fclose(fp);

		delete[] splines;
		return;
	}


	// initialize to black
	glClearColor(0.f, 0.f, 0.f, 1.f);

	glClear(GL_COLOR_BUFFER_BIT);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	dh = (max_offset - min_offset) / (num_strokes + 1);



#if USE_GLSL_SPLINE == 1

	glEnable(GL_TEXTURE_2D);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	if(param.embossing_type == 0) {
		GLSLManager::SetShader("Height.vert", "", "Height0.frag");
	}
	else if(param.embossing_type == 1) {
		GLSLManager::SetShader("Height.vert", "", "Height1.frag");
	}

	program_id = GLSLManager::GetCurrentProgram();
	jitter_scale_loc = glGetUniformLocation(program_id, "jitter_scale");
	jitter_freq_loc = glGetUniformLocation(program_id, "jitter_freq");


	// end of initialization...
	D_TOCK(mt, d_cnt_idx++)


	for(int i = 0; i < max_layers; i++) {
		D_TICK(mt, d_cnt_idx)

		stroke_itr = layers[i]->strokes.rbegin();
		spline_itr = splines[i].rbegin();
		spline_end = splines[i].rend();

		// I assume that all strokes are within a certain range such that x:[0, width - 1] and y:[0, height - 1]
		if(param.embossing_type == 0) {
			for(; spline_itr != spline_end; spline_itr++, stroke_itr++) {
				pp = &(fsn[(int)(*stroke_itr)->x + (int)(*stroke_itr)->y * width].preset);
				offset = min_offset + (cnt++) * dh;

				glUniform1f(jitter_scale_loc, pp->post_process_val3);
				glUniform1f(jitter_freq_loc, pp->post_process_val4);

				(*spline_itr)->setHeightMapTexture(hm_ptr[(*stroke_itr)->jitter_id]);		
				(*spline_itr)->drawHeight(&offset);
			}
		}
		else if(param.embossing_type == 1) {
			for(; spline_itr != spline_end; spline_itr++, stroke_itr++) {
				pp = &(fsn[(int)(*stroke_itr)->x + (int)(*stroke_itr)->y * width].preset);

				glUniform1f(jitter_scale_loc, pp->post_process_val3);
				glUniform1f(jitter_freq_loc, pp->post_process_val4);

				(*spline_itr)->setHeightMapTexture(hm_ptr[(*stroke_itr)->jitter_id]);
				(*spline_itr)->drawHeight(NULL);
			}
		}

		D_TOCK(mt, d_cnt_idx++)
	}


	// cleanup
	D_TICK(mt, d_cnt_idx)

	GLSLManager::SetShader();
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);

	D_TOCK(mt, d_cnt_idx++)

#else

	setUpMultiTexturePostProcess();

	// enable vertex array and 4 tex. coord. arrays
	glClientActiveTexture(GL_TEXTURE0);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE1);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE2);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	glEnableClientState(GL_VERTEX_ARRAY);


	// end of initialization...
	D_TOCK(mt, d_cnt_idx++)


	if(param.embossing_type == 0) {
		for(int i = 0; i < max_layers; i++) {
			D_TICK(mt, d_cnt_idx)

			stroke_itr = layers[i]->strokes.rbegin();
			spline_itr = splines[i].rbegin();
			spline_end = splines[i].rend();

			for(; spline_itr != spline_end; spline_itr++, stroke_itr++) {
				pp = &(fsn[(int)(*stroke_itr)->x + (int)(*stroke_itr)->y * width].preset);
				offset = min_offset + (cnt++) * dh;

				(*spline_itr)->setHeightMapTexture(hm_ptr[(*stroke_itr)->jitter_id]);
				(*spline_itr)->setHeightMapTexHeightScale(pp->post_process_val3 * frac);
				(*spline_itr)->setHeightMapTexCoordScale(pp->post_process_val4);
				(*spline_itr)->drawHeight(&offset);
			}

			D_TOCK(mt, d_cnt_idx++)
		}
	}
	else {
		for(int i = 0; i < max_layers; i++) {
			D_TICK(mt, d_cnt_idx)

			stroke_itr = layers[i]->strokes.rbegin();
			spline_itr = splines[i].rbegin();
			spline_end = splines[i].rend();

			for(; spline_itr != spline_end; spline_itr++, stroke_itr++) {
				pp = &(fsn[(int)(*stroke_itr)->x + (int)(*stroke_itr)->y * width].preset);

				(*spline_itr)->setHeightMapTexture(hm_ptr[(*stroke_itr)->jitter_id]);
				(*spline_itr)->setHeightMapTexHeightScale(pp->post_process_val3 * frac);
				(*spline_itr)->setHeightMapTexCoordScale(pp->post_process_val4);
				(*spline_itr)->drawHeight(NULL);
			}

			D_TOCK(mt, d_cnt_idx++)
		}
	}


	// clean up
	D_TICK(mt, d_cnt_idx)

	// disable
	glDisableClientState(GL_VERTEX_ARRAY);

	glClientActiveTexture(GL_TEXTURE2);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE1);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE0);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	cleanUpMultiTexturePostProcess();

	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);

	D_TOCK(mt, d_cnt_idx++)

#endif

	glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, painted_image);

	//My_SSS_SavePPM(width, height, painted_image, "heightmap.ppm");

	delete[] splines;




#if D_SAVESTAT_SPLINE == 1
	int num_layers = 0;
	string path(VisView::getFullPath());
	path.append("\\Spline_stat.txt");
	FILE *d_fp = fopen(path.c_str(), "a");


	if(param.style->getNumRegions() == 1) {
		num_layers = (param.style->getStyleNode())[0].preset.num_layers;
	}
	else {
		num_layers = param.style->getMaxLayers();
	}


	d_cnt_idx = 0;
	fprintf(d_fp, "Heightmap rendering:                                     \tTime(sec)\n");
	fprintf(d_fp, "Initialize: (enable GL states)                           \t%.3f\n\n", D_TIME(mt, d_cnt_idx, timer_freq)); d_cnt_idx++;

	for(int i = 0; i < num_layers; i++) {
		fprintf(d_fp, "Layer %d\n", i);
		fprintf(d_fp, "  Render:                                                \t%.3f\n", D_TIME(mt, d_cnt_idx, timer_freq)); d_cnt_idx++;
		fprintf(d_fp, "\n");
	}

	fprintf(d_fp, "Clean up (disable GL states):                            \t%.3f\n", D_TIME(mt, d_cnt_idx, timer_freq)); d_cnt_idx++;
	fprintf(d_fp, "---------------------------------------------------------------\n");
	fprintf(d_fp, "Render Time:                                             \t%.3f\n", (double)(mt[1][d_cnt_idx-1]-mt[0][0]) / (double)timer_freq);
	fprintf(d_fp, "\n\n");
	fclose(d_fp);

#endif

	D_FIN_TIMING(mt)
}


const unsigned char* SplineRenderer::getInitialImage(int layer) {
	float scale = 1.f;
	Spline_SimpleLayer *current_layer = NULL;
	spline_list tmp_splines;
	
	RenderTexture *prev_rt;

	if(0 <= layer && layer < max_layers) {
		current_layer = layers[layer];

		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glPushClientAttrib(GL_CLIENT_ALL_ATTRIB_BITS);

		prev_rt = GLSLManager::GetCurrentRenderTarget();
		
		GLSLManager::SetRenderTarget(tmp_target_RT);

		// just return a pointer or regenerate
		if(initial_image_regeneration[layer] == false) {
			for(int i = 0; i < layer; i++) scale *= .5f;

			glClearColor(1.f, 1.f, 1.f, 0.f);		// set the white canvas
			glClear(GL_COLOR_BUFFER_BIT);
			glEnable(GL_TEXTURE_2D);
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);

			traceStrokes(current_layer->strokes, tmp_splines, scale, 0, true);
			renderTexturedStrokes(tmp_splines);
			glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, current_layer->initial_stroke_image);

			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisableClientState(GL_VERTEX_ARRAY);
			glDisable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);

			spline_factory.dump(tmp_splines);		// reuse

			initial_image_regeneration[layer] = true;
		}


		// restore
		if(prev_rt != NULL) {
			GLSLManager::SetRenderTarget(prev_rt);
		}
		else {
			GLSLManager::SetRenderTarget();
		}

		glPopClientAttrib();
		glPopAttrib();

		return current_layer->initial_stroke_image;
	}
	else {
		return NULL;
	}
}


#undef USE_GLSL_SPLINE
#undef D_SET_COUNTER_VAR
#undef D_UPDATE_COUNTER_VAR
#undef D_INIT_TIMING
#undef D_TICK
#undef D_TOCK
#undef D_TIME
#undef D_FIN_TIMING
#undef D_USE_DEPTH_SPLINE
