#pragma once
#include <list>


using namespace std;


class SplineStroke {
public:
	enum SPLINE_TYPE {
		CATMULL_ROM, 
		UNIFORM_B_SPLINE, 
	};


	SplineStroke(float x0, float y0, float stroke_size, float stroke_color[4], float init_angle = 0, int density = 1) {
		this->stroke_size = stroke_size;
		this->density = density;
		length = 0;
		squared_length = 0;
		num_spline_points = 0;
		tesselated = false;
		num_ctrl_points = 1;
		spline_points = NULL;
		param_list = NULL;
		threshold = init_threshold;
		first_angle = init_angle;
		id = -1;

		//xc = x0;
		//yc = y0;
		init_x = x0;
		init_y = y0;
		f_xc = x0;
		f_yc = y0;
		b_xc = x0;
		b_yc = y0;

		ctrl_x.push_back(x0);
		ctrl_y.push_back(y0);
		
		color[0] = stroke_color[0];
		color[1] = stroke_color[1];
		color[2] = stroke_color[2];
		color[3] = stroke_color[3];

		param_list_capacity = 0;
		spline_points_capacity = 0;

		createParamList();
		setSplineType(UNIFORM_B_SPLINE);
	}


	virtual ~SplineStroke() {
		delete[] spline_points;
		delete[] param_list;
	}


	virtual void reset(float x0, float y0, float stroke_size, float stroke_color[4], float init_angle = 0, int density = 1);


	inline int getNumCtrlPoints() { return num_ctrl_points; }
	inline int getNumSplinePoints() { return num_ctrl_points + (num_ctrl_points - 1) * density; }

	inline void getColor(float *color) { 
		color[0] = this->color[0]; 
		color[1] = this->color[1]; 
		color[2] = this->color[2]; 
		color[3] = this->color[3]; 
	}

	inline float getStrokeSize() { return stroke_size; }
	inline float getStrokeLength() { return length; }
	inline float getStrokeSquaredLength() { return squared_length; }

	inline void setThreshold(float th) { threshold = (th < 0) ? 0 : th; }
	inline float getThreshold() { return threshold; }

	inline float getInitX() { return init_x; }
	inline float getInitY() { return init_y; }

	void addFront(float x, float y);
	void addBack(float x, float y);
	virtual bool tesselate(bool add_ends = false);

	inline void setID(int val) { id = val; }
	inline int getID() { return id; }

	bool isTesselated() { return tesselated; }

	// library dependent... (OpenGL, OpenCV, etc...)
	virtual bool draw(void *params) = 0;


	void setSplineType(SplineStroke::SPLINE_TYPE type);
	SplineStroke::SPLINE_TYPE getSplineType() { return type; }


protected:
	void createParamList() {
		int dp2 = density + 2;
		float factor = 1.f / (dp2 - 1);


		if(param_list == NULL) {
			param_list = new float[dp2];
			param_list_capacity = dp2;
		}
		else if(param_list_capacity < dp2) {
			delete[] param_list;
			param_list = new float[dp2];
			param_list_capacity = dp2;
		}

		for(int i = 0; i < dp2; i++) {
			param_list[i] = i * factor;
		}
	}

	int density;
	int num_ctrl_points;
	int num_spline_points;
	bool tesselated;
	float color[4];
	//float xc, yc;
	float init_x, init_y;
	float f_xc, f_yc;
	float b_xc, b_yc;
	float stroke_size;
	float squared_length, length;
	list<float> ctrl_x, ctrl_y;
	float *spline_points;
	float *param_list;
	float threshold;
	float first_angle;
	int id;


	int param_list_capacity;
	int spline_points_capacity;


	SPLINE_TYPE type;

	static const float init_threshold;
};

