#pragma once
#include "corona.h"
#include <string>
#include <gl\gl.h>
#pragma comment(lib, "corona.lib")


using namespace std;


class StrokeMapManager {
public:
	static GLuint makeStrokeShapeTexture(string filename) {
		GLuint tex_handle = 0;
		int idx = 0;
		int offset = 0;
		int num_pixels;
		int num_elems;
		unsigned char *buffer, *ptr;
		corona::Image* image = corona::OpenImage(filename.c_str());

		if(image != NULL) {
			num_pixels = image->getWidth() * image->getHeight();
			num_elems = 4 * num_pixels;
			buffer = new unsigned char[num_elems];
			ptr = (unsigned char*)image->getPixels();
			offset = getOffset(*image);

			if(offset > 0) {
				for(int i = 0; i < num_pixels; i++) {
					buffer[idx++] = 255;
					buffer[idx++] = 255;
					buffer[idx++] = 255;
					buffer[idx++] = ptr[offset * i];
				}

				tex_handle = uploadTexure(image->getWidth(), image->getHeight(), buffer);
			}

			delete image;
			delete[] buffer;
		}

		return tex_handle;
	}

	static GLuint makeStrokeHeightmapTexture(string filename) {
		GLuint tex_handle = 0;
		int idx = 0;
		int offset = 0;
		int num_pixels;
		int num_elems;
		unsigned char *buffer, *ptr;
		corona::Image* image = corona::OpenImage(filename.c_str());

		if(image != NULL) {
			num_pixels = image->getWidth() * image->getHeight();
			num_elems = 4 * num_pixels;
			buffer = new unsigned char[num_elems];
			ptr = (unsigned char*)image->getPixels();
			offset = getOffset(*image);

			if(offset > 0) {
				for(int i = 0; i < num_pixels; i++) {
					buffer[idx++] = ptr[offset * i];
					buffer[idx++] = ptr[offset * i];
					buffer[idx++] = ptr[offset * i];
					buffer[idx++] = 255;
				}

				tex_handle = uploadTexure(image->getWidth(), image->getHeight(), buffer);
			}

			delete image;
			delete[] buffer;
		}

		return tex_handle;
	}


private:
	static GLuint uploadTexure(int w, int h, unsigned char *buffer, bool build_mipmap = false) {
		GLuint tex_handle = 0;
		int status = 0;

		glGenTextures(1, &tex_handle);
		glBindTexture(GL_TEXTURE_2D, tex_handle);

		// ... tried different texparameter for min_filter, but none of them isn't in my favor.
		if(build_mipmap == true) {
			status = gluBuild2DMipmaps(GL_TEXTURE_2D, 4, w, h, GL_RGBA, GL_UNSIGNED_BYTE, buffer);

			if(status != 0) {
				FILE *fp = fopen("blah.txt", "w");
				fprintf(fp, "StrokeMapManager::uploadTexture(): %s\r\n", gluErrorString(glGetError()));
				fclose(fp);
			}

			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		}
		else {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		}

		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glBindTexture(GL_TEXTURE_2D, 0);

		return tex_handle;
	}


	static int getOffset(corona::Image &image) {
		switch(image.getFormat()) {
			case corona::PF_DONTCARE:
				return -1;
				break;

			case corona::PF_R8G8B8A8:
			case corona::PF_B8G8R8A8:
				return 4;
				break;

			case corona::PF_R8G8B8:
			case corona::PF_B8G8R8:
				return 3;
				break;

			case corona::PF_I8:
				return 1;
				break;

			default:
				return -1;
				break;
		}
	}

};
