#pragma warning(disable: 4482)
#pragma once

#ifndef __VisView_H__
#define __VisView_H__

#include "glew.h"
//#include <GL/gl.h>
#include <GL/glu.h>
#include <vector>
#include <utility>
#include "image_pyramid.h"
#include "tendes_plane.h"
#include "cv.h"
#include "cvaux.h"
#include "StrokeParams.h"
#include "PaintingPreset.h"
#include "PathMaker.h"
#include "PaintingStyleField.h"
#include "PaintingRegion.h"
#include "EdgeFieldSampler.h"
#include "EdgeFieldManipulator.h"
#include "EdgeFieldConstructor.h"
#include "GLSLManager.h"
#include "MFCTEST4Dlg.h"
#include "IBFVRenderer.h"
#include "SplineRenderer.h"
#include "Segmentation.h"
#include "RegionLabel.h"
#include "RegionTransformation.h"

using namespace std;

class CMFCTEST4Dlg;


class VisView : public CWnd {
private:
	CWnd    *hWnd;
	CMFCTEST4Dlg *parent;
	HDC     hdc;
	HGLRC   hrc;
	int     m_nPixelFormat;
	CRect   m_originalRect;


	static TCHAR full_path[MAX_PATH];
	Polyhedron *poly;

public:
	/* Constructor/destructor, methods for initialization... */
	VisView();
	virtual ~VisView();
	void Create(CRect rect, CWnd *parent, LPSTR filename);

	bool resize(int width, int height);

	static const string getFullPath() { return string(full_path); }


private:
	void initializeData();
	void initializeBuffer();
	void initializeRenderer();
	void finalizeBuffer();
	void finalizeRenderer();
	void updateMeshVisibility();


public:
	float testparam_f;



	/* Function to grab an image currently displayed... NOTE: image origin is at lower left */
public:
	//origin... 0: top-left origin, 1: bottom-left origin, assumed that the image buffer is supplied as a part of parameters...
	void grabImage(unsigned char **img, int origin = 1);



	/* We recursively call OnTimer() for drawing... sometimes we need to temporarily stop the timer... */
private:
	UINT_PTR m_unpTimer;

public:
	bool getDisplayTimerStatus();
	void stopDisplayTimer();
	void startDisplayTimer();



	/* OpenGL related */
public:
	void oglInitialize();
	void initGLView();


	/* Visibility function */
private:
	bool run_once;
	int num_visible_triangles;
	int *visible_tri_idx;
	float *visible_tri_vertex_coords;
	float *visible_tri_vertex_tex_coords[8];

protected:
	void getVisibilitForTriangles(int &num_visibles, int * (&tri_idx));


	/* Mouse mode */
private:
	/*
		-2 = no action, 
		-1 = down,
		0 = zoom, 
		1 = rotate x, 
		2 = rotate y, 
		3 = tranlate x, 
		4 = translate y, 
		5 = cull near, 
		6 = cull far 
	*/
	int mouse_mode;

public:
	static const enum MOUSE_MODE {
		NO_MOUSE_ACTION = -2, 
		DOWN, 
		ZOOM, 
		ROTATE_X, 
		ROTATE_Y, 
		TRANSLATE_X, 
		TRANSLATE_Y, 
		CULL_NEAR, 
		CULL_FAR, 
	};



	/* Mouse operations & conversion between client and drawing coordinates */
private:
	int pick_object(int x, int y);
	int processHits(GLint hits, GLuint buffer[]);
	Triangle *d_tri_spec;



	/* Dimensions, Pixels */
private:
	int win_width, win_height;
	int num_win_pixels;

	int img_width, img_height;
	int num_img_pixels;

public:
	inline int getWinWidth() { return win_width; }
	inline int getWinHeight() { return win_height; }
	inline int getImgWidth() { return img_width; }
	inline int getImgHeight() { return img_height; }



	/* Action state for field design */
private:
	int act_state;
	int old_act_state;

public:
	static const enum ACT_STATE {
		NO_ACTION = -1, 
		ADD_SINK, 
		ADD_SOURCE, 
		ADD_SADDLE, 
		ADD_CCW_CENTER, 
		ADD_CW_CENTER, 
		EDIT_ELEMENT, 
		ADD_REGULAR_ELEMENT, 
		REMOVE_ELEMENT, 
		SELECT_ELEMENT, 
		SINGULARITY_PAIR_CANCELLATION, 
		MOVE_SINGULARITY, 
		ADD_DIPOLE1, 
		ADD_DIPOLE2, 
		ADD_MONKEY_SADDLE, 
	};

	void setActionState(VisView::ACT_STATE action);




	/*
		 0: Store each flow image independently
		 1: V+x, V-x, V+y, V-y share the same flow image
		 2: Show composite of V+x and V+y (flow image not shared)
		 3: Show composite of V+x and V+y (flow image shared)
		 4: Display compsite of V+x and V-x (flow image not shared)
		 5: Display compsite of V+x and V-x (flow image shared)
		 6: Show only V+x
		 7: Show only V-x
		 8: Show only V+y
		 9: Show only V-y
		10: Display compsite of V+y and V-y (flow image not shared)
		11: Display compsite of V-x and V-y (flow image not shared)
	*/
private:
	int image_composition_method;
	bool copy_back;
	int coords[4];
	int num_enabled_warped_images;

public:
	void setImageSynthesisMethod(int type);
	void makeBlendMap();


	/* Field design */
private:
	int edit_mode;   // -1 = none, 0 = translation, 1 = rotation, 2-9 = stretch from 8 positions

	int current_se_id, current_re_id;
	int current_analytical_cp_id, last_analytical_cp_id;


	// used to temporarily save design elements from Polyhedron class
	SingularElementList *selist;
	RegularElementList *relist;

	EdgeFieldManipulator field_manipulator;

public:
	void setGlobalRotationAngle(double degree);
	void global_rotate_field();
	void global_reflect_field();
	void smoothClosedField();

	void create_user_spec(int win_x, int win_y);

	void undo();
	void redo();

	void loadControlFromFile(const char *filename);
	void saveControlToFile(const char *filename);
	void clearSelectedRegion();

	void saveDesignElements();
	void setDesignElements();

	void resetPropagation() { if(painting_style_field && poly) painting_style_field->propagateDesignElements(poly->singularelements, poly->regularelements); }


	/* Preprocessing for IBFV such as image processing, sampling, smoothing, etc... */
public:

	// assume that origin is at left-top
	FieldSampleImagePyramid *input;

private:
	int img_smoothing_for_edgefield_iterations;
	const static int init_img_smoothing_for_edgefield_iterations = 7;

	int edge_field_smoothing_iterations;
	const static int init_edge_field_smoothing_iterations = 40;


	unsigned char *original_image;			// store the current original image
	float *blend_map;						// weight map for tensor field images
	int num_sample;							// number of samples

	unsigned char* segmentation_image;       //store the segmentation image

	unsigned char *stroke_ref_img;			// optional image palette to select stroke colors. 
	bool use_stroke_ref_img;


	EdgeFieldConstructor field_constructor;
	EdgeFieldSampler field_sampler;


public:
	void setImgSmoothingIterationsForEdgeField(int num_itr);
	int getImgSmoothingIterationsForEdgeField();

	void setEdgeFieldSmoothingIterations(int num_itr);
	int getEdgeFieldSmoothingIterations();

	void setBaseImage(unsigned char *img, int w, int h);								// default image to use...
	void setStrokeRefImage(const unsigned char *img = NULL, int w = 0, int h = 0);		// optional to replace a reference image for stroke color.
	void enableStrokeImageInput(bool val) { use_stroke_ref_img = val; }
	bool isStrokeImageInputEnabled() { return use_stroke_ref_img; }

	bool computeFieldFromImage(FieldSampleImage *fsi = NULL);
	void sampleFieldToFSI(FieldSampleImage *fsi = NULL);
	void constructFieldInformation();

	bool loadEdgeField(string &filename);		// load tensor based edge field data from file. 
	bool saveEdgeField(string &filename);		// load tensor based edge field data from file. 


	/* Displaying */
public:

	static const enum DISPLAY_MODE {
		VECTOR_EDGE_FIELD, 
		TENSOR_EDGE_FIELD, 
		PAINTING, 
		BLACK, 
		INPUT_IMAGE, 
		BLENDMAP, 
		SEGMENTATION,
		LABELIMAGE,
		DATIMAGE,
		STROKE_REF_IMAGE, 
	};

	static const enum TENSOR_VIS_MODE {
		MAJOR, 
		MINOR, 
		MAJOR_MINOR, 
	};

	static const enum EDGE_FIELD_TYPE {
		VECTOR, 
		TENSOR, 
	};

private:
	DISPLAY_MODE current_display_mode;
	TENSOR_VIS_MODE current_tensor_vis_mode;

	int desired_width;
	double render_time;
	__int64 timer_freq;
	double seg_time;

	void display_vector_knobs();
	void display_tensor_knobs();
	void display_triangles();


public:
	void setDisplayMode(VisView::DISPLAY_MODE mode);
	VisView::DISPLAY_MODE getDisplayMode();

	void setTensorVisMode(VisView::TENSOR_VIS_MODE mode);
	VisView::TENSOR_VIS_MODE getTensorVisMode();
	VisView::EDGE_FIELD_TYPE getEdgeFieldType();

	void display(bool print_stat = false);
	void display_knobs();
	double getRenderTime();




	/* Painting method, texture generation seting... */
public:
	static const enum COLOR_COMPOSITION {
		ALPHA_BLEND, 
		OVERWRITE, 
	};

	static const int num_circular_stroke_coords = 24;

	double strong_vertex_ratio;		// from 0 to 1


private:
	static const enum PAINTING_STEP {
		PAINTING_DONE, 
		IMAGE_PAINTING, 
		POST_PROCESSING, 
	};

	PAINTING_STEP current_painting_step;



	static const int init_NPN_size = 128;
	static const double init_max_displacemnt;		// 255./NPIX
	static const int max_supported_layers = 3;

	int NPN_size;

	int warp_iterations_counter;
	int num_warping_for_paint;
	int num_warping_for_vis;

	COLOR_COMPOSITION color_composition_method_for_paint;
	COLOR_COMPOSITION color_composition_method_for_vis;

	int blend_alpha_for_paint;
	int blend_alpha_for_vis;



	int stroke_regen_mode;			// 0: regenerate, 1: retain current set of strokes, and add new ones if necessary


	RenderTexture *offscreen;					// offscreen-rendering
	RenderTexture *offscreen_blendmap;			// offscreen for a blend map
	QuadVBO *quad;								// render off-screen texture on screen


	unsigned char *noise_pattern_image;


	// stroke-base rendering components
	GLint viewport[4];
	GLdouble modelview[16];
	GLdouble projection[16];



	PaintingStyleField *painting_style_field;
	PaintingRegion *selected_region;

	unsigned char *pixels_all[8];
	unsigned char *vis_image;
	unsigned char *painted_image;
	unsigned char *painted_image_copy;


	void makeRandMap();

	void renderVectorField();
	void renderTensorField();
	void display_one_flow_image(int which_coordinates, int num_itr = 1);

	void displayPainting();			// Function for painting
	void displayEdgeField();		// Function for visualizing edge field

	static void alphablendColor(unsigned char **images, unsigned char blend_val, int width, int height, unsigned char *output);
	void weightblendColor(unsigned char **images, float *weight_map, int width, int height, unsigned char *output, int op = 0);
	void makeNoisePatterns();		// black-white noise map for visualization


public:

	void initializeRendering();
	void setBackgroundPattern();

	void setNPNSize(int npn_size);
	int getNPNsize();


	void setVisIterations(int num_itr);
	void setPaintIterations(int num_itr);
	int getVisIterations();
	int getPaintIterations();

	void setVisDisplacement(double dp);
	double getVisDisplacement();

	void setVisBlendAlpha(int alpha);
	void setPaintBlendAlpha(int alpha);
	int getVisBlendAlpha();
	int getPaintBlendAlpha();

	void setVisColorCompositionMethod(VisView::COLOR_COMPOSITION composition);
	void setPaintColorCompositionMethod(VisView::COLOR_COMPOSITION composition);
	VisView::COLOR_COMPOSITION getVisColorCompositionMethod();
	VisView::COLOR_COMPOSITION getPaintColorCompositionMethod();



	void setInterpolateWeight(double weight);
	double getInterpolateWeight();
	void interpolateField();
	void resetField();							// take all design elements out of the current edge field
	void clearField();							// zeroes field on underline mesh

	void setStrokeRegenerationMode(int mode);

	// change values of stroke parameters on the '1st' layer
	void setStrokeParams(StrokeParams stroke_params);
	StrokeParams getStrokeParams();

	bool setNumEnabledLayers(int new_num_layers);
	int getNumEnabledLayers();
	int getMaxSupportedLayers();

	void setAreaErrorThreshold(float val);
	float getAreaErrorThreshold();

	size_t getNumMultiStyles();

	void setTraceMode(PathMaker::MODE new_mode);
	PathMaker::MODE getTraceMode();


	size_t getNumStrokes(int layer_idx);
	void clearStrokes(int layer_idx);
	void shiftStrokes(int layer_idx, float *shift, int width, int height);


	double dt;
	int dt_itr;
	void updatePaintingStyleField();
	void updateTexCoordsOnMesh();
	void copyTexCoordsList();


	void syncRegionInfoOnMesh();
	void updateRegionInfoOnMesh();


	/* Post Processing */
public:
	static const enum PDV_TYPE {
		EDGE_DARKING, 
		HIGH_FREQ, 
		LOW_FREQ, 
		PAPER_NOISE, 
	};


private:
	unsigned char *height_map;
	double *post_processing_map;
	double *turbulance_map;
	double *pigment_map;
	double *paper_texture_map;

	int post_processing_operation;		// 0: pigment dispersion, 1: embossing

	const static int init_post_processing_operation = 0;

	bool enable_post_processing;
	bool update_post_processing_map;


	void createStaticTexturesForWaterColorEffect();
	void createEdgeDarkingMap();
	void createHeightMap();
	void applyPigmentDensityVariation(VisView::PDV_TYPE type);
	void applyLighting();


public:	
	void enablePostProcessing(bool val);
	bool getPostProcessingEnabled();

	void setPostProcessingOperation(int type);
	int getPostProcessingOperation();

	void applyPostProcessing();

	void setPostProcessingParam(int param, float val);
	float getPostProcessingParam(int param);




	/* Toggle to display additional features */
private:
	bool show_singularities;
	bool show_separatrices;
	bool show_control_knobs;
	bool show_multistyle_regions;

public:
	void setShowSeparatrices(bool val);
	void setShowSingularities(bool val);
	void setShowControlKnobs(bool val);
	void setShowMultiStyleRegions(bool val);
	void setShowCurves(bool val);

	bool getShowSeparatrices();
	bool getShowSingularities();
	bool getShowControlKnobs();
	bool getShowMultiStyleRegions();
	bool getShowCurves();



private:
	int current_time;
	int num_frames;
	//////////////////////////////////////////////////////////////////////////////////////////////
	unsigned int regionID1,regionID2;
	////////////////////////////////////////////////////////////////////////////////////////////////

public:
	int getNumFrames();
	void setNumFrames(int num_frames, const string &file_path = string());
	int getCurrentTime();
	void setCurrentTime(int time);
	int findNextKey();
	int findPrevKey();
	bool removeKey();
	void setKey(PaintingPreset p);
	void updateKey();
	PaintingPreset getKey();

	void setKeyInterpolationFunc(TransitionalPaintingPreset::INTERPOLATION_TYPE type);
	TransitionalPaintingPreset::INTERPOLATION_TYPE getKeyInterpolationFunc();

	int findNextDEKey();
	int findPrevDEKey();


	/* Segmenting Region... */
private:
	bool show_curves;
	float curve_color[4];
	bool draw_traces;
	typedef vector< pair<int, int> > PATH;
	PATH traced_points;
	vector<PATH*> curves;
	PathMaker path_maker;


	void displayTracedPoints();
	void displayMultiStyleRegions();
	void displayMultiStyleCurves();


public:

	void resetPaintingRegions(bool retain_style = false);
	void setRegion(unsigned char *binary_map, int width, int height);
	PaintingRegion *getSelectedRegion() {return selected_region; }


private:
	CvPoint2D32f *optflow;

public:
	void setOptFlowForVis(float *data, int width, int height);
	void displayOptFlowArrows();




	/* Some debug-purpose flags */
public:
	bool dbg_show_layer;				// flag to show a layer separately
	int dbg_show_layer_idx;				// which layer to show
	bool dbg_show_only_initial_stroke_image;
	bool dbg_show_optflow;

	bool dbg_output_stroke_map;
	bool dbg_save_advection_image;
	bool dbg_save_stroke_placement_mask;



	/* Color ordering, IBFV comparison... */
private:
	int stroke_order;			// 0: random, 1: RGB, 2: HSV
	bool partial_ordering;		// partial ordering if true, entire ordering if false

	int embossing_type;			// 0: legacy (Hertzmann), 1: intensity

public:
	void setStrokeOrder(int type);
	void setEmbossingType(int type);
	void setOrderingRange(bool partial_or_whole);
	int getStrokeOrder() { return stroke_order; }
	int getEmbossingType() { return embossing_type; }
	bool getOrderingRange() { return partial_ordering; }

	/* Renderer */
public:
	static const enum RENDERER {
		IBFV, 
		SPLINE, 
	};

	static const enum RENDERING_OPTION {
		IBFV_USE_GL_NEAERST_IN_WARP, 
		IBFV_BGFG, 
	};

private:
	RENDERER renderer;

	IBFVRenderer *ibfv_renderer;
	IBFVParam ibfv_param;
	SplineRenderer *spline_renderer;
	SplineParam spline_param;
	ILayeredStrokePaintRenderer *current_renderer;
	IPaintingParam *current_painting_param;

	void setRendererParameter();

public:
	VisView::RENDERER getRenderingMethod() { return renderer; }
	void setRenderingMethod(VisView::RENDERER method);

	void setRenderingOption_b(RENDERING_OPTION option, bool val);
	bool getRenderingOption_b(RENDERING_OPTION option);
	void setRenderingOption_uc(RENDERING_OPTION option, unsigned char val);
	unsigned char getRenderingOption_uc(RENDERING_OPTION option);

public:
	afx_msg void OnPaint();
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg void OnSize(UINT nType, int cx, int cy);
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg int OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnMButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnMButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint point);
	DECLARE_MESSAGE_MAP()
	public:
	void getSegmentation(float Spatial, int Range, int MinSize);

public:
	Segmentation seg_img;
public:
	bool segmImageDefined;
public:
	void setSegmentationRegion(void);
	int flag_merge;
	int current_id;
	int flag_click_merge;
	clock_t time1;
	clock_t merge_time1;
	clock_t merge_time2;

public:
	void setSegImage(unsigned char* img, int w, int h);


private:
	// Data file (Transformation)
	RegionTransformation dat;

	// video segmentation
	int VideoSegmentation;


public:
	void loadDat(const char *filename);
	void unloadDat();
	void setTransformation();
};

#endif /* __VisView_H__ */
