#include "StdAfx.h"
#include "image_pyramid.h"


bool FieldSampleImage::blur(PaintingStyleField &psf) {
	int num_elems;
	int h_1 = h - 1;
	bool no_edgefield_defined = true;
	list< pair<int, int> > *points;
	list< pair<int, int> >::iterator itr_list;
	map<int, PaintingRegion*>::iterator itr_map;
	TensorInterpolatable *edge_field;


	// do nothing...
	if(diffuse == false) return true;


	// dimension check
	if(psf.getWidth() != w || psf.getHeight() != h) return false;


	/**
	CString msg;
	for(itr_map = psf.getBegin(); itr_map != psf.getEnd(); itr_map++) {
		msg.Format("%d: %s\r\n", itr_map->second->getID(), itr_map->second->getDiffusionFlag() ? "T" : "F");
		MessageBox(0, msg, "", MB_OK);
	}
	/**/


	/**
	// test to see which region has to be diffused
	FILE *fp = fopen("test.txt", "w");
	for(itr_map = psf.getBegin(); itr_map != psf.getEnd(); itr_map++) {
		if((*itr_map).second->getDiffusionFlag() == false) {
			fprintf(fp, "region %d: %c\n", itr_map->second->getID(), 'F');
		}
		else {
			fprintf(fp, "region %d: %c\n", itr_map->second->getID(), 'T');
		}
	}
	fclose(fp);
	/**/


	for(itr_map = psf.getBegin(); itr_map != psf.getEnd(); itr_map++) {
		if((*itr_map).second->getDiffusionFlag() == false) {
			no_edgefield_defined = false;
			break;
		}
	}


	// no interpolation is necessary if no edge field is fixed.
	if(no_edgefield_defined == true) return true;

	/**
	CString str;
	for(itr_map = psf.getBegin(); itr_map != psf.getEnd(); itr_map++) {
		if((*itr_map).second->getDiffusionFlag() == false) {
			str.Format("%d: false", itr_map->second->getID());
			MessageBox(0, str, "", MB_OK);
		}
		else {
			str.Format("%d: true", itr_map->second->getID());
			MessageBox(0, str, "", MB_OK);
		}
	}
	/**/


	num_elems = w * h;
	edge_field = new TensorInterpolatable[num_elems];


	// apply mask
	for(itr_map = psf.getBegin(); itr_map != psf.getEnd(); itr_map++) {
		points = &((*itr_map).second->points);

		if((*itr_map).second->getDiffusionFlag() == false) {
			// make sure of using different indices due to different origin location
			for(itr_list = points->begin(); itr_list != points->end(); itr_list++) {
				edge_field[itr_list->first + w * (h_1 - itr_list->second)].mask = false;
			}
		}
		else {
			for(itr_list = points->begin(); itr_list != points->end(); itr_list++) {
				edge_field[itr_list->first + w * (h_1 - itr_list->second)].mask = true;
			}
		}
	}

	for(int i = 0; i < num_elems; i++) {
		if(edge_field[i].mask == true) {
			fld[i].ten_fld.set(0, 0, 0, 0);
		}

		edge_field[i].pack(fld[i].ten_fld);
	}


	// blur (heat diffusion)
	ef_blur.setConvergenceCriteria(1.0e-4f);
	diffuse_stepsize = 10.f;
	ef_blur.blur(edge_field, w, h, diffuse_stepsize, diffuse_num_itr);


	for(int i = 0; i < num_elems; i++) {
		edge_field[i].unpack(fld[i].ten_fld);
	}

	delete[] edge_field;

	return true;
}
