/*

Data structure and programs for working with images

Eugene Zhang 2003

*/

#pragma once
#include "stdafx.h"
#include "ppm.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <fstream>
#include <iostream>
#include <string>


#define ASCII_HT 9
#define ASCII_NL 10
#define ASCII_CR 13
#define ASCII_SPACE 32


using namespace std;



int My_SSS_SavePPM(int w, int h, unsigned char* data, LPCTSTR fName) {
	int num_rows = 3 * w;
	FILE* f_p = _tfopen(fName, _T("wb"));

	fprintf(f_p, "P6\n%i %i\n255\n", w, h);

	for(int i = h - 1; i >= 0; i--){
		fwrite(&(data[num_rows * i]), sizeof(char), num_rows, f_p);
		fflush(0);
	}

	fclose(f_p);
	return 0;
}


int My_SSS_ReadPPM(int *w, int *h, unsigned char** data, LPCTSTR fName) {
	int dyn_range;			// usually 255, so no use
	int dum_i, idx;
	string dum_str;
	ifstream ifs;

	_tprintf(_T("File name: %s\n"), fName);

	ifs.open(fName, ifstream::binary);
	*w = -1;
	*h = -1;
	dyn_range = -1;


	while(ifs.eof() == false) {
		dum_i = ifs.peek();

		if(dum_i == 'P') {
			ifs >> dum_str;						// magic number
		}
		else if(dum_i == '#') {					// comment
			while(ifs.peek() != ASCII_NL) {
				ifs.get();
			}

			ifs.get();
		}
		else if(dum_i == ASCII_HT || dum_i == ASCII_NL || dum_i == ASCII_CR || dum_i == ASCII_SPACE) {
			ifs.get();

			if(dyn_range > 0) break;
		}
		else {
			if(*w < 0) {
				ifs >> *w >> *h;
			}
			else {
				ifs >> dyn_range;
			}
		}
	}

	*data = new unsigned char[*w * *h * 3];

	if(ifs.eof() == false) {
		for(int i = 0; i < *h; i++) {
			for(int j = 0; j < *w; j++) {
				idx = j + i * *w;

				(*data)[3 * idx + 0] = (unsigned char)ifs.get();
				(*data)[3 * idx + 1] = (unsigned char)ifs.get();
				(*data)[3 * idx + 2] = (unsigned char)ifs.get();
			}
		}
	}

	ifs.close();
	return 0;
}


void push_pull(unsigned char *src, unsigned char **dst, int orig_w, int orig_h, int desired_w, int desired_h) {
	int i, j, k, l;
	int n_horizontal=0;
	int pyramid_level = (int)(log(desired_w+1.0)/log(2.0));
	int pyramid_width = desired_w;
	int pyramid_height = 2*desired_w - 1;

	unsigned char *pixels = new unsigned char[3 * desired_w * desired_h];
	unsigned char *handled = new unsigned char[pyramid_height * pyramid_width];
	double *final_value = new double[3 * pyramid_height * pyramid_width];

	for (i=0; i<pyramid_height*pyramid_width; i++)
		handled[i] = 0;
	for (i=0; i<3*pyramid_height*pyramid_width; i++)
		final_value[i] = 0.0;

	// let i be width, j be height
	for (i=0; i<desired_h; i++) {
		for (j=0; j<desired_w; j++) {
			int handle_index = j+i*desired_w, index = handle_index*3;
			if (i<orig_h && j<orig_w) {
				int orig_index = (j+i*orig_w)*3;
				final_value[index] = src[orig_index];
				final_value[index+1] = src[orig_index+1];
				final_value[index+2] = src[orig_index+2];
				handled[handle_index] = 1;
			} else {
				handled[handle_index] = 0;
			}
		}
	}

	for (i=0; i<3*pyramid_width*pyramid_height; i++) {
		if (final_value[i] > 255.0)
			final_value[i] = 255.0;
		else if (final_value[i] < 0.0)
			final_value[i] = 0.0;
	}

	for (i=0; i<desired_h; i++) {
		for (j=0; j<desired_w; j++) {
			int index = (j+i*desired_w)*3;

			pixels[index] = (unsigned char)final_value[index];
			pixels[index+1] = (unsigned char)final_value[index+1];
			pixels[index+2] = (unsigned char)final_value[index+2];

		}
	}

	int prev_row_start, row_start = 0, prev_row_height, row_height = desired_h;
	int tx, ty, the_pos;
	int px[4], py[4], pos[4];
	for (k=pyramid_level-1; k>=0; k--) {
		prev_row_start = row_start;
		prev_row_height = row_height;
		row_start += row_height;
		row_height /= 2;
		for (i=0; i<row_height; i++) {
			for (j=0; j<row_height; j++) {
				tx = j;
				ty = row_start + i;
				the_pos = ty*desired_h+tx;
				px[0] = 2*j;
				py[0] = prev_row_start + 2*i;
				px[1] = px[0]+1;
				py[1] = py[0];
				px[2] = px[0];
				py[2] = py[0]+1;
				px[3] = px[0]+1;
				py[3] = py[0]+1;
				for (l=0; l<4; l++) {
					pos[l] = py[l]*desired_h+px[l];
					if (handled[pos[l]] == 1) {
						handled[the_pos]++;
						final_value[the_pos*3] += final_value[pos[l]*3];
						final_value[the_pos*3+1] += final_value[pos[l]*3+1];
						final_value[the_pos*3+2] += final_value[pos[l]*3+2];
					} 
				}
				if (handled[the_pos] > 0) {
					final_value[the_pos*3] /= handled[the_pos];
					final_value[the_pos*3+1] /= handled[the_pos];
					final_value[the_pos*3+2] /= handled[the_pos];
					handled[the_pos] = 1;
				}
			}
		}
	}

	for (k=0; k<pyramid_level; k++) {
		prev_row_start = row_start;
		prev_row_height = row_height;
		row_height *= 2;
		row_start -= row_height;
		for (i=0; i<row_height/2; i++) {
			for (j=0; j<row_height/2; j++) {
				tx = j;
				ty = prev_row_start + i;
				the_pos = ty*desired_h+tx;
				px[0] = 2*j;
				py[0] = row_start + 2*i;
				px[1] = px[0]+1;
				py[1] = py[0];
				px[2] = px[0];
				py[2] = py[0]+1;
				px[3] = px[0]+1;
				py[3] = py[0]+1;
				for (l=0; l<4; l++) {
					pos[l] = py[l]*desired_h+px[l];
					if (handled[pos[l]] == 0) {
						final_value[pos[l]*3] = final_value[the_pos*3];
						final_value[pos[l]*3+1] = final_value[the_pos*3+1];
						final_value[pos[l]*3+2] = final_value[the_pos*3+2];
					}
				}
			}
		}
	}

	for (i=0; i<3*pyramid_width*pyramid_height; i++) {
		if (final_value[i] > 255.0)
			final_value[i] = 255.0;
		else if (final_value[i] < 0.0)
			final_value[i] = 0.0;
	}

	for (i=0; i<desired_h; i++) {
		for (j=0; j<desired_w; j++) {
			int index = (j+i*desired_w)*3;

			pixels[index] = (unsigned char)final_value[index];
			pixels[index+1] = (unsigned char)final_value[index+1];
			pixels[index+2] = (unsigned char)final_value[index+2];

		}
	}

	delete[] final_value;
	delete[] handled;

	(*dst) = pixels;
}
