/*

Header file for skip-lists.

*/

#ifndef __SKIPLIST_H__
#define __SKIPLIST_H__

/* one element in a skip-list */

typedef struct Skip_element {		/* an element in a skip-list */
  char *data;				/* data for the element */
  int level;				/* how many forward pointers */
  struct Skip_element **forward;	/* forward pointer(s) */
} Skip_element;

/* a skip-list */

class SkipList {            /* skip-list (variant on linked-list) */
private:
  int (*compare)(char *, char *);     /* comparison procedure */
  int num_elements;                   /* number of elements in list */
  int max_level;                      /* maximum pointer level */
public:
  Skip_element head;                  /* first pointers to all levels */
  SkipList(int (*compare)(char *, char *), int);
  ~SkipList();
  void add(char *);
  int remove(char *);
  char *find(char *);
  char *first();
  char *last();
  void print(void (print_data)(char *));
  int count();
};

/* traverse a skip-list in sorted order */

#define list_traverse(list,elem) \
  for (elem = (list)->head.forward[0]; elem != NULL; elem = elem->forward[0])


#endif /* __SKIPLIST_H__ */

